# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------
"""Implements blockwise processing methods in the frequency domain.

.. autosummary::
    :toctree: generated/

    RFFT
    IRFFT
    AutoPowerSpectra
    CrossPowerSpectra
    FFTSpectra
"""

from warnings import warn

import numpy as np
from scipy import fft
from traits.api import Bool, CArray, Enum, Instance, Int, Property, Union, cached_property

# acoular imports
from .base import SamplesGenerator, SpectraGenerator, SpectraOut, TimeOut
from .deprecation import deprecated_alias
from .fastFuncs import calcCSM
from .internal import digest
from .process import SamplesBuffer
from .spectra import BaseSpectra


@deprecated_alias({'numfreqs': 'num_freqs', 'numsamples': 'num_samples'}, read_only=True)
class RFFT(BaseSpectra, SpectraOut):
    """Provides the one-sided Fast Fourier Transform (FFT) for real-valued multichannel time data.

    The FFT is calculated block-wise, i.e. the input data is divided into blocks of length
    :attr:`block_size` and the FFT is calculated for each block. Optionally, a window function
    can be applied to the data before the FFT calculation via the :attr:`window` attribute.
    """

    #: Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    #: Number of workers to use for the FFT calculation. If negative values are used,
    #: all available logical CPUs will be considered (``scipy.fft.rfft`` implementation wraps around
    #: from ``os.cpu_count()``).
    #: Default is `None` (handled by scipy)
    workers = Union(Int(), None, default_value=None, desc='number of workers to use')

    #: Scaling method, either 'amplitude', 'energy' or :code:`none`.
    #: Default is :code:`none`.
    #: 'energy': compensates for the energy loss due to truncation of the FFT result. The resulting
    #: one-sided spectrum is multiplied by 2.0, except for the DC and Nyquist frequency.
    #: 'amplitude': scales the one-sided spectrum so that the amplitude of discrete tones does not
    #: depend on the block size.
    scaling = Enum('none', 'energy', 'amplitude')

    #: block size of the FFT. Default is 1024.
    block_size = Property()

    #: Number of frequencies in the output.
    num_freqs = Property(depends_on=['_block_size'])

    #: Number of snapshots in the output.
    num_samples = Property(depends_on=['source.num_samples', '_block_size'])

    #: 1-D array of FFT sample frequencies.
    freqs = Property()

    # internal block size variable
    _block_size = Int(1024, desc='block size of the FFT')

    # internal identifier
    digest = Property(depends_on=['source.digest', 'scaling', 'precision', '_block_size', 'window', 'overlap'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_num_freqs(self):
        return int(self.block_size / 2 + 1)

    @cached_property
    def _get_num_samples(self):
        if self.source.num_samples >= 0:
            return int(np.floor(self.source.num_samples / self.block_size))
        return -1

    def _get_block_size(self):
        return self._block_size

    def _set_block_size(self, value):
        if value % 2 != 0:
            msg = 'Block size must be even'
            raise ValueError(msg)
        self._block_size = value

    def _scale(self, data, scaling_value):
        """Corrects the energy of the one-sided FFT data."""
        if self.scaling == 'amplitude' or self.scaling == 'energy':
            data[1:-1] *= 2.0
        data *= scaling_value
        return data

    def _get_freqs(self):
        """Return the Discrete Fourier Transform sample frequencies.

        Returns
        -------
        f : ndarray
            1-D Array of length *block_size/2+1* containing the sample frequencies.

        """
        if self.source is not None:
            return abs(fft.fftfreq(self.block_size, 1.0 / self.source.sample_freq)[: int(self.block_size / 2 + 1)])
        return np.array([])

    def result(self, num=1):
        """Python generator that yields the output block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the number of multi-channel spectra (i.e. snapshots) per block
            returned by the generator.

        Returns
        -------
        Spectra block of shape (num, :attr:`num_channels` * :attr:`num_freqs`).
            The last block may be shorter than num.

        """
        wind = self.window_(self.block_size)
        if self.scaling == 'none' or self.scaling == 'energy':  # only compensate for the window
            svalue = 1 / np.sqrt(np.dot(wind, wind) / self.block_size)
        elif self.scaling == 'amplitude':  # compensates for the window and the energy loss
            svalue = 1 / wind.sum()
        wind = wind[:, np.newaxis]
        fftdata = np.zeros((num, self.num_channels * self.num_freqs), dtype=self.precision)
        j = 0
        for i, data in enumerate(self._get_source_data()):  # yields one block of time data
            j = i % num
            fftdata[j] = self._scale(
                fft.rfft(data * wind, n=self.block_size, axis=0, workers=self.workers).astype(self.precision),
                scaling_value=svalue,
            ).reshape(-1)
            if j == num - 1:
                yield fftdata
        if j < num - 1:  # yield remaining fft spectra
            yield fftdata[: j + 1]


@deprecated_alias({'numsamples': 'num_samples'}, read_only=True)
class IRFFT(TimeOut):
    """Calculates the inverse Fast Fourier Transform (IFFT) for one-sided multi-channel spectra."""

    #: Data source; :class:`~acoular.base.SpectraGenerator` or derived object.
    source = Instance(SpectraGenerator)

    #: Number of workers to use for the FFT calculation. If negative values are used,
    #: all available logical CPUs will be considered (``scipy.fft.rfft`` implementation wraps around
    #: from ``os.cpu_count()``).
    #: Default is `None` (handled by scipy)
    workers = Union(Int(), None, default_value=None, desc='number of workers to use')

    #: The floating-number-precision of the resulting time signals, corresponding to numpy dtypes.
    #: Default is 64 bit.
    precision = Enum('float64', 'float32', desc='precision of the time signal after the ifft')

    #: Number of time samples in the output.
    num_samples = Property(depends_on=['source.num_samples', 'source._block_size'])

    # internal time signal buffer to handle arbitrary output block sizes
    _buffer = CArray(desc='signal buffer')

    # internal identifier
    digest = Property(depends_on=['source.digest', 'scaling', 'precision', '_block_size', 'window', 'overlap'])

    def _get_num_samples(self):
        if self.source.num_samples >= 0:
            return int(self.source.num_samples * self.source.block_size)
        return -1

    @cached_property
    def _get_digest(self):
        return digest(self)

    def _validate(self):
        if not self.source.block_size or self.source.block_size < 0:
            msg = (
                f'Source of class {self.__class__.__name__} has an unknown blocksize: {self.source.block_size}.'
                'This is likely due to incomplete spectral data from which the inverse FFT cannot be calculated.'
            )
            raise ValueError(msg)
        if (self.source.num_freqs - 1) * 2 != self.source.block_size:
            msg = (
                f'Block size must be 2*(num_freqs-1) but is {self.source.block_size}.'
                'This is likely due to incomplete spectral data from which the inverse FFT cannot be calculated.'
            )
            raise ValueError(msg)
        if self.source.block_size % 2 != 0:
            msg = f'Block size must be even but is {self.source.block_size}.'
            raise ValueError(msg)

    def result(self, num):
        """Python generator that yields the output block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block). The last block may be shorter than num.

        Yields
        ------
        numpy.ndarray
            Yields blocks of shape (num, num_channels).
        """
        self._validate()
        bs = self.source.block_size
        if num != bs:
            buffer_length = (int(np.ceil(num / bs)) + 1) * bs
            buffer = SamplesBuffer(source=self, source_num=bs, length=buffer_length, dtype=self.precision)
            yield from buffer.result(num)
        else:
            for spectra in self.source.result(1):
                yield fft.irfft(
                    spectra.reshape(self.source.num_freqs, self.num_channels), n=num, axis=0, workers=self.workers
                )


class AutoPowerSpectra(SpectraOut):
    """Calculates the real-valued auto-power spectra."""

    #: Data source; :class:`~acoular.base.SpectraGenerator` or derived object.
    source = Instance(SpectraGenerator)

    #: Scaling method, either 'power' or 'psd' (Power Spectral Density).
    #: Only relevant if the source is a :class:`~acoular.fprocess.FreqInOut` object.
    scaling = Enum('power', 'psd')

    #: Determines if the spectra yielded by the source are single-sided spectra.
    single_sided = Bool(True, desc='single sided spectrum')

    #: The floating-number-precision of entries, corresponding to numpy dtypes. Default is 64 bit.
    precision = Enum('float64', 'float32', desc='floating-number-precision')

    # internal identifier
    digest = Property(depends_on=['source.digest', 'precision', 'scaling', 'single_sided'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    def _get_scaling_value(self):
        scale = 1 / self.block_size**2
        if self.single_sided:
            scale *= 2
        if self.scaling == 'psd':
            scale *= self.block_size * self.source.sample_freq
        return scale

    def result(self, num=1):
        """Python generator that yields the real-valued auto-power spectra.

        Parameters
        ----------
        num : integer
            This parameter defines the number of snapshots within each output data block.

        Yields
        ------
        numpy.ndarray
            Yields blocks of shape (num, num_channels * num_freqs).
            The last block may be shorter than num.

        """
        scale = self._get_scaling_value()
        for temp in self.source.result(num):
            yield ((temp * temp.conjugate()).real * scale).astype(self.precision)


@deprecated_alias({'numchannels': 'num_channels'}, read_only=True)
class CrossPowerSpectra(AutoPowerSpectra):
    """Calculates the complex-valued auto- and cross-power spectra.

    Receives the complex-valued spectra from the source and returns the cross-spectral matrix (CSM)
    in a flattened representation (i.e. the auto- and cross-power spectra are concatenated along the
    last axis). If :attr:`calc_mode` is 'full', the full CSM is calculated, if 'upper', only the
    upper triangle is calculated.
    """

    #: Data source; :class:`~acoular.base.SpectraGenerator` or derived object.
    source = Instance(SpectraGenerator)

    #: The floating-number-precision of entries of csm, eigenvalues and
    #: eigenvectors, corresponding to numpy dtypes. Default is 64 bit.
    precision = Enum('complex128', 'complex64', desc='precision of the fft')

    #: Calculation mode, either 'full' or 'upper'.
    #: 'full' calculates the full cross-spectral matrix, 'upper' calculates
    # only the upper triangle. Default is 'full'.
    calc_mode = Enum('full', 'upper', 'lower', desc='calculation mode')

    #: Number of channels in output. If :attr:`calc_mode` is 'full', then
    #: :attr:`num_channels` is :math:`n^2`, where :math:`n` is the number of
    #: channels in the input. If :attr:`calc_mode` is 'upper', then
    #: :attr:`num_channels` is :math:`n + n(n-1)/2`.
    num_channels = Property(depends_on=['source.num_channels'])

    # internal identifier
    digest = Property(depends_on=['source.digest', 'precision', 'scaling', 'single_sided', 'calc_mode'])

    @cached_property
    def _get_num_channels(self):
        n = self.source.num_channels
        return n**2 if self.calc_mode == 'full' else int(n + n * (n - 1) / 2)

    @cached_property
    def _get_digest(self):
        return digest(self)

    def result(self, num=1):
        """Python generator that yields the output block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Yields
        ------
        numpy.ndarray
            Yields blocks of shape (num, num_channels * num_freqs).
        """
        nc_src = self.source.num_channels
        nc = self.num_channels
        nf = self.num_freqs
        scale = self._get_scaling_value()

        csm_flat = np.zeros((num, nc * nf), dtype=self.precision)
        csm_upper = np.zeros((nf, nc_src, nc_src), dtype=self.precision)
        for data in self.source.result(num):
            for i in range(data.shape[0]):
                calcCSM(csm_upper, data[i].astype(self.precision).reshape(nf, nc_src))
                if self.calc_mode == 'full':
                    csm_lower = csm_upper.conj().transpose(0, 2, 1)
                    [np.fill_diagonal(csm_lower[cntFreq, :, :], 0) for cntFreq in range(csm_lower.shape[0])]
                    csm_flat[i] = (csm_lower + csm_upper).reshape(-1)
                elif self.calc_mode == 'upper':
                    csm_flat[i] = csm_upper[:, :nc].reshape(-1)
                else:  # lower
                    csm_lower = csm_upper.conj().transpose(0, 2, 1)
                    csm_flat[i] = csm_lower[:, :nc].reshape(-1)
                csm_upper[...] = 0  # calcCSM adds cumulative
            yield csm_flat[: i + 1] * scale


class FFTSpectra(RFFT):
    """Provides the one-sided Fast Fourier Transform (FFT) for multichannel time data.

    Alias for :class:`~acoular.fprocess.RFFT`.

    .. deprecated:: 24.10
        Using :class:`~acoular.fprocess.FFTSpectra` is deprecated and will be removed in Acoular
        version 25.07. Use :class:`~acoular.fprocess.RFFT` instead.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warn(
            'Using FFTSpectra is deprecated and will be removed in Acoular version 25.07. Use class RFFT instead.',
            DeprecationWarning,
            stacklevel=2,
        )



# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------
"""Estimation of power spectra and related tools.

.. autosummary::
    :toctree: generated/

    BaseSpectra
    PowerSpectra
    PowerSpectraImport
"""

from abc import abstractmethod
from warnings import warn

from numpy import (
    arange,
    array,
    bartlett,
    blackman,
    dot,
    empty,
    fill_diagonal,
    hamming,
    hanning,
    imag,
    linalg,
    ndarray,
    newaxis,
    ones,
    real,
    searchsorted,
    sum,  # noqa A004
    zeros,
)
from scipy import fft
from traits.api import (
    ABCHasStrictTraits,
    Bool,
    CArray,
    Delegate,
    Enum,
    Float,
    Instance,
    Int,
    Map,
    Property,
    Union,
    cached_property,
    property_depends_on,
)

# acoular imports
from .base import SamplesGenerator
from .configuration import config
from .deprecation import deprecated_alias
from .fastFuncs import calcCSM
from .h5cache import H5cache
from .h5files import H5CacheFileBase
from .internal import digest
from .tools.utils import find_basename


@deprecated_alias({'numchannels': 'num_channels'}, read_only=True)
class BaseSpectra(ABCHasStrictTraits):
    #: Data source; :class:`~acoular.sources.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    #: Sampling frequency of output signal, as given by :attr:`source`.
    sample_freq = Delegate('source')

    #: Number of time data channels
    num_channels = Delegate('source')

    #: Window function for FFT, one of:
    #:   * 'Rectangular' (default)
    #:   * 'Hanning'
    #:   * 'Hamming'
    #:   * 'Bartlett'
    #:   * 'Blackman'
    window = Map(
        {'Rectangular': ones, 'Hanning': hanning, 'Hamming': hamming, 'Bartlett': bartlett, 'Blackman': blackman},
        default_value='Rectangular',
        desc='type of window for FFT',
    )

    #: Overlap factor for averaging: 'None'(default), '50%', '75%', '87.5%'.
    overlap = Map({'None': 1, '50%': 2, '75%': 4, '87.5%': 8}, default_value='None', desc='overlap of FFT blocks')

    #: FFT block size, one of: 128, 256, 512, 1024, 2048 ... 65536,
    #: defaults to 1024.
    block_size = Enum(
        1024,
        128,
        256,
        512,
        2048,
        4096,
        8192,
        16384,
        32768,
        65536,
        desc='number of samples per FFT block',
    )

    #: The floating-number-precision of the resulting spectra, corresponding to numpy dtypes.
    #: Default is 'complex128'.
    precision = Enum('complex128', 'complex64', desc='precision of the fft')

    # internal identifier
    digest = Property(depends_on=['precision', 'block_size', 'window', 'overlap'])

    @abstractmethod
    def _get_digest(self):
        """Return internal identifier."""

    def fftfreq(self):
        """Return the Discrete Fourier Transform sample frequencies.

        Returns
        -------
        f : ndarray
            Array of length *block_size/2+1* containing the sample frequencies.

        """
        if self.source is not None:
            return abs(fft.fftfreq(self.block_size, 1.0 / self.source.sample_freq)[: int(self.block_size / 2 + 1)])
        return None

    # generator that yields the time data blocks for every channel (with optional overlap)
    def _get_source_data(self):
        bs = self.block_size
        temp = empty((2 * bs, self.num_channels))
        pos = bs
        posinc = bs / self.overlap_
        for data_block in self.source.result(bs):
            ns = data_block.shape[0]
            temp[bs : bs + ns] = data_block  # fill from right
            while pos + bs <= bs + ns:
                yield temp[int(pos) : int(pos + bs)]
                pos += posinc
            else:
                temp[0:bs] = temp[bs:]  # copy to left
                pos -= bs


class PowerSpectra(BaseSpectra):
    """Provides the cross spectral matrix of multichannel time data
     and its eigen-decomposition.

    This class includes the efficient calculation of the full cross spectral
    matrix using the Welch method with windows and overlap (:cite:`Welch1967`). It also contains
    the CSM's eigenvalues and eigenvectors and additional properties.

    The result is computed only when needed, that is when the :attr:`csm`,
    :attr:`eva`, or :attr:`eve` attributes are actually read.
    Any change in the input data or parameters leads to a new calculation,
    again triggered when an attribute is read. The result may be
    cached on disk in HDF5 files and need not to be recomputed during
    subsequent program runs with identical input data and parameters. The
    input data is taken to be identical if the source has identical parameters
    and the same file name in case of that the data is read from a file.
    """

    #: Data source; :class:`~acoular.sources.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    # Shadow trait, should not be set directly, for internal use.
    _ind_low = Int(1, desc='index of lowest frequency line')

    # Shadow trait, should not be set directly, for internal use.
    _ind_high = Union(Int(-1), None, desc='index of highest frequency line')

    #: Index of lowest frequency line to compute, integer, defaults to 1,
    #: is used only by objects that fetch the csm, PowerSpectra computes every
    #: frequency line.
    ind_low = Property(_ind_low, desc='index of lowest frequency line')

    #: Index of highest frequency line to compute, integer,
    #: defaults to -1 (last possible line for default block_size).
    ind_high = Property(_ind_high, desc='index of lowest frequency line')

    # Stores the set lower frequency, for internal use, should not be set directly.
    _freqlc = Float(0)

    # Stores the set higher frequency, for internal use, should not be set directly.
    _freqhc = Union(Float(0), None)

    # Saves whether the user set indices or frequencies last, for internal use only,
    # not to be set directly, if True (default), indices are used for setting
    # the freq_range interval.
    _index_set_last = Bool(True)

    #: Flag, if true (default), the result is cached in h5 files and need not
    #: to be recomputed during subsequent program runs.
    cached = Bool(True, desc='cached flag')

    #: Number of FFT blocks to average, readonly
    #: (set from block_size and overlap).
    num_blocks = Property(desc='overall number of FFT blocks')

    #: 2-element array with the lowest and highest frequency. If set,
    #: will overwrite :attr:`_freqlc` and :attr:`_freqhc` according to
    #: the range.
    #: The freq_range interval will be the smallest discrete frequency
    #: inside the half-open interval [_freqlc, _freqhc[ and the smallest
    #: upper frequency outside of the interval.
    #: If user chooses the higher frequency larger than the max frequency,
    #: the max frequency will be the upper bound.
    freq_range = Property(desc='frequency range')

    #: Array with a sequence of indices for all frequencies
    #: between :attr:`ind_low` and :attr:`ind_high` within the result, readonly.
    indices = Property(desc='index range')

    #: Name of the cache file without extension, readonly.
    basename = Property(depends_on=['source.digest'], desc='basename for cache file')

    #: The cross spectral matrix,
    #: (number of frequencies, num_channels, num_channels) array of complex;
    #: readonly.
    csm = Property(desc='cross spectral matrix')

    #: Eigenvalues of the cross spectral matrix as an
    #: (number of frequencies) array of floats, readonly.
    eva = Property(desc='eigenvalues of cross spectral matrix')

    #: Eigenvectors of the cross spectral matrix as an
    #: (number of frequencies, num_channels, num_channels) array of floats,
    #: readonly.
    eve = Property(desc='eigenvectors of cross spectral matrix')

    # internal identifier
    digest = Property(
        depends_on=['source.digest', 'block_size', 'window', 'overlap', 'precision'],
    )

    # hdf5 cache file
    h5f = Instance(H5CacheFileBase, transient=True)

    @property_depends_on(['source.num_samples', 'block_size', 'overlap'])
    def _get_num_blocks(self):
        return self.overlap_ * self.source.num_samples / self.block_size - self.overlap_ + 1

    @property_depends_on(['source.sample_freq', 'block_size', 'ind_low', 'ind_high'])
    def _get_freq_range(self):
        fftfreq = self.fftfreq()
        if fftfreq is not None:
            if self._ind_high is None:
                return array([fftfreq[self.ind_low], None])
            return fftfreq[[self.ind_low, self.ind_high]]
        return None

    def _set_freq_range(self, freq_range):  # by setting this the user sets _freqlc and _freqhc
        self._index_set_last = False
        self._freqlc = freq_range[0]
        self._freqhc = freq_range[1]

    @property_depends_on(['source.sample_freq', 'block_size', '_ind_low', '_freqlc'])
    def _get_ind_low(self):
        fftfreq = self.fftfreq()
        if fftfreq is not None:
            if self._index_set_last:
                return min(self._ind_low, fftfreq.shape[0] - 1)
            return searchsorted(fftfreq[:-1], self._freqlc)
        return None

    @property_depends_on(['source.sample_freq', 'block_size', '_ind_high', '_freqhc'])
    def _get_ind_high(self):
        fftfreq = self.fftfreq()
        if fftfreq is not None:
            if self._index_set_last:
                if self._ind_high is None:
                    return None
                return min(self._ind_high, fftfreq.shape[0] - 1)
            if self._freqhc is None:
                return None
            return searchsorted(fftfreq[:-1], self._freqhc)
        return None

    def _set_ind_high(self, ind_high):  # by setting this the user sets the lower index
        self._index_set_last = True
        self._ind_high = ind_high

    def _set_ind_low(self, ind_low):  # by setting this the user sets the higher index
        self._index_set_last = True
        self._ind_low = ind_low

    @property_depends_on(['block_size', 'ind_low', 'ind_high'])
    def _get_indices(self):
        fftfreq = self.fftfreq()
        if fftfreq is not None:
            try:
                indices = arange(fftfreq.shape[0], dtype=int)
                if self.ind_high is None:
                    return indices[self.ind_low :]
                return indices[self.ind_low : self.ind_high]
            except IndexError:
                return range(0)
        return None

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_basename(self):
        return find_basename(self.source, alternative_basename=self.source.__class__.__name__ + self.source.digest)

    def calc_csm(self):
        """Csm calculation."""
        t = self.source
        wind = self.window_(self.block_size)
        weight = dot(wind, wind)
        wind = wind[newaxis, :].swapaxes(0, 1)
        numfreq = int(self.block_size / 2 + 1)
        csm_shape = (numfreq, t.num_channels, t.num_channels)
        csm_upper = zeros(csm_shape, dtype=self.precision)
        # get time data blockwise
        for data in self._get_source_data():
            ft = fft.rfft(data * wind, None, 0).astype(self.precision)
            calcCSM(csm_upper, ft)  # only upper triangular part of matrix is calculated (for speed reasons)
        # create the full csm matrix via transposing and complex conj.
        csm_lower = csm_upper.conj().transpose(0, 2, 1)
        [fill_diagonal(csm_lower[cntFreq, :, :], 0) for cntFreq in range(csm_lower.shape[0])]
        csm = csm_lower + csm_upper
        # onesided spectrum: multiplication by 2.0=sqrt(2)^2
        return csm * (2.0 / self.block_size / weight / self.num_blocks)

    def calc_ev(self):
        """Eigenvalues / eigenvectors calculation."""
        if self.precision == 'complex128':
            eva_dtype = 'float64'
        elif self.precision == 'complex64':
            eva_dtype = 'float32'
        #        csm = self.csm #trigger calculation
        csm_shape = self.csm.shape
        eva = empty(csm_shape[0:2], dtype=eva_dtype)
        eve = empty(csm_shape, dtype=self.precision)
        for i in range(csm_shape[0]):
            (eva[i], eve[i]) = linalg.eigh(self.csm[i])
        return (eva, eve)

    def calc_eva(self):
        """Calculates eigenvalues of csm."""
        return self.calc_ev()[0]

    def calc_eve(self):
        """Calculates eigenvectors of csm."""
        return self.calc_ev()[1]

    def _get_filecache(self, traitname):
        """Function handles result caching of csm, eigenvectors and eigenvalues
        calculation depending on global/local caching behaviour.
        """
        if traitname == 'csm':
            func = self.calc_csm
            numfreq = int(self.block_size / 2 + 1)
            shape = (numfreq, self.source.num_channels, self.source.num_channels)
            precision = self.precision
        elif traitname == 'eva':
            func = self.calc_eva
            shape = self.csm.shape[0:2]
            if self.precision == 'complex128':
                precision = 'float64'
            elif self.precision == 'complex64':
                precision = 'float32'
        elif traitname == 'eve':
            func = self.calc_eve
            shape = self.csm.shape
            precision = self.precision

        H5cache.get_cache_file(self, self.basename)
        if not self.h5f:  # in case of global caching readonly
            return func()

        nodename = traitname + '_' + self.digest
        if config.global_caching == 'overwrite' and self.h5f.is_cached(nodename):
            # print("remove existing node",nodename)
            self.h5f.remove_data(nodename)  # remove old data before writing in overwrite mode

        if not self.h5f.is_cached(nodename):
            if config.global_caching == 'readonly':
                return func()
            #            print("create array, data not cached for",nodename)
            self.h5f.create_compressible_array(nodename, shape, precision)

        ac = self.h5f.get_data_by_reference(nodename)
        if ac[:].sum() == 0:  # only initialized
            #            print("write {} to:".format(traitname),nodename)
            ac[:] = func()
            self.h5f.flush()
        return ac

    @property_depends_on(['digest'])
    def _get_csm(self):
        """Main work is done here:
        Cross spectral matrix is either loaded from cache file or
        calculated and then additionally stored into cache.
        """
        if config.global_caching == 'none' or (config.global_caching == 'individual' and self.cached is False):
            return self.calc_csm()
        return self._get_filecache('csm')

    @property_depends_on(['digest'])
    def _get_eva(self):
        """Eigenvalues of cross spectral matrix are either loaded from cache file or
        calculated and then additionally stored into cache.
        """
        if config.global_caching == 'none' or (config.global_caching == 'individual' and self.cached is False):
            return self.calc_eva()
        return self._get_filecache('eva')

    @property_depends_on(['digest'])
    def _get_eve(self):
        """Eigenvectors of cross spectral matrix are either loaded from cache file or
        calculated and then additionally stored into cache.
        """
        if config.global_caching == 'none' or (config.global_caching == 'individual' and self.cached is False):
            return self.calc_eve()
        return self._get_filecache('eve')

    def synthetic_ev(self, freq, num=0):
        """Return synthesized frequency band values of the eigenvalues.

        Parameters
        ----------
        freq : float
            Band center frequency for which to return the results.
        num : integer
            Controls the width of the frequency bands considered; defaults to
            3 (third-octave band).

            ===  =====================
            num  frequency band width
            ===  =====================
            0    single frequency line
            1    octave band
            3    third-octave band
            n    1/n-octave band
            ===  =====================

        Returns
        -------
        float
            Synthesized frequency band value of the eigenvalues (the sum of
            all values that are contained in the band).

        """
        f = self.fftfreq()
        if num == 0:
            # single frequency line
            return self.eva[searchsorted(f, freq)]
        f1 = searchsorted(f, freq * 2.0 ** (-0.5 / num))
        f2 = searchsorted(f, freq * 2.0 ** (0.5 / num))
        if f1 == f2:
            return self.eva[f1]
        return sum(self.eva[f1:f2], 0)


class PowerSpectraImport(PowerSpectra):
    """Provides a dummy class for using pre-calculated cross-spectral
    matrices.

    This class does not calculate the cross-spectral matrix. Instead,
    the user can inject one or multiple existing CSMs by setting the
    :attr:`csm` attribute. This can be useful when algorithms shall be
    evaluated with existing CSM matrices.
    The frequency or frequencies contained by the CSM must be set via the
    attr:`frequencies` attribute. The attr:`num_channels` attributes
    is determined on the basis of the CSM shape.
    In contrast to the PowerSpectra object, the attributes
    :attr:`sample_freq`, :attr:`source`, :attr:`block_size`, :attr:`window`,
    :attr:`overlap`, :attr:`cached`, and :attr:`num_blocks`
    have no functionality.
    """

    #: The cross spectral matrix,
    #: (number of frequencies, num_channels, num_channels) array of complex;
    csm = Property(desc='cross spectral matrix')

    #: frequencies included in the cross-spectral matrix in ascending order.
    #: Compound trait that accepts arguments of type list, array, and float
    frequencies = Union(CArray, Float, desc='frequencies included in the cross-spectral matrix')

    #: Number of time data channels
    num_channels = Property(depends_on=['digest'])

    source = Enum(None, desc='PowerSpectraImport cannot consume time data')

    # Sampling frequency of the signal, defaults to None
    sample_freq = Enum(None, desc='sampling frequency')

    block_size = Enum(None, desc='PowerSpectraImport does not operate on blocks of time data')

    window = Enum(None, desc='PowerSpectraImport does not perform windowing')

    overlap = Enum(None, desc='PowerSpectraImport does not consume time data')

    cached = Enum(False, desc='PowerSpectraImport has no caching capabilities')

    num_blocks = Enum(None, desc='PowerSpectraImport cannot determine the number of blocks')

    # Shadow trait, should not be set directly, for internal use.
    _ind_low = Int(0, desc='index of lowest frequency line')

    # Shadow trait, should not be set directly, for internal use.
    _ind_high = Union(None, Int, desc='index of highest frequency line')

    # internal identifier
    digest = Property(depends_on=['_csmsum'])

    #: Name of the cache file without extension, readonly.
    basename = Property(depends_on=['digest'], desc='basename for cache file')

    # csm shadow trait, only for internal use.
    _csm = CArray()

    # CSM checksum to trigger digest calculation, only for internal use.
    _csmsum = Float()

    def _get_basename(self):
        return 'csm_import_' + self.digest

    @cached_property
    def _get_digest(self):
        return digest(self)

    def _get_num_channels(self):
        return self.csm.shape[1]

    def _get_csm(self):
        return self._csm

    def _set_csm(self, csm):
        if (len(csm.shape) != 3) or (csm.shape[1] != csm.shape[2]):
            msg = 'The cross spectral matrix must have the following shape: \
            (number of frequencies, num_channels, num_channels)!'
            raise ValueError(msg)
        self._csmsum = real(self._csm).sum() + (imag(self._csm) ** 2).sum()  # to trigger new digest creation
        self._csm = csm

    @property_depends_on(['digest'])
    def _get_eva(self):
        """Eigenvalues of cross spectral matrix are either loaded from cache file or
        calculated and then additionally stored into cache.
        """
        return self.calc_eva()

    @property_depends_on(['digest'])
    def _get_eve(self):
        """Eigenvectors of cross spectral matrix are either loaded from cache file or
        calculated and then additionally stored into cache.
        """
        return self.calc_eve()

    def fftfreq(self):
        """Return the Discrete Fourier Transform sample frequencies.

        Returns
        -------
        f : ndarray
            Array containing the frequencies.

        """
        if isinstance(self.frequencies, float):
            return array([self.frequencies])
        if isinstance(self.frequencies, ndarray):
            return self.frequencies
        if self.frequencies is None:
            warn('No frequencies defined for PowerSpectraImport object!', stacklevel=1)
        return self.frequencies



# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------
"""Implements the definition of trajectories.

.. autosummary::
    :toctree: generated/

    Trajectory
"""

# imports from other packages
from numpy import arange, array, r_, sort
from scipy.interpolate import splev, splprep
from traits.api import Dict, Float, HasStrictTraits, Property, Tuple, cached_property, property_depends_on

# acoular imports
from .internal import digest


class Trajectory(HasStrictTraits):
    """Describes a trajectory from sampled points.

    Based on a discrete number of points in space and time, a
    continuous trajectory is calculated using spline interpolation
    of positions between samples.
    """

    #: Dictionary that assigns discrete time instants (keys) to
    #: sampled `(x, y, z)` positions along the trajectory (values).
    points = Dict(
        key_trait=Float,
        value_trait=Tuple(Float, Float, Float),
        desc='sampled positions along the trajectory',
    )

    #: Tuple of the start and end time, is set automatically
    #: (depending on :attr:`points`).
    interval = Property()
    # t_min, t_max tuple

    #: Spline data, internal use.
    tck = Property()

    # internal identifier
    digest = Property(depends_on=['points[]'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    @property_depends_on(['points[]'])
    def _get_interval(self):
        return sort(list(self.points.keys()))[r_[0, -1]]

    @property_depends_on(['points[]'])
    def _get_tck(self):
        t = sort(list(self.points.keys()))
        xp = array([self.points[i] for i in t]).T
        k = min(3, len(self.points) - 1)
        tcku = splprep(xp, u=t, s=0, k=k)
        return tcku[0]

    def location(self, t, der=0):
        """Returns the positions for one or more instants in time.

        Parameters
        ----------
        t : array of floats
            Instances in time to calculate the positions at.
        der : integer
            The order of derivative of the spline to compute, defaults to 0.

        Returns
        -------
        (x, y, z) : tuple with arrays of floats
            Positions at the given times; `x`, `y` and `z` have the same shape as `t`.

        """
        return splev(t, self.tck, der)

    def traj(self, t_start, t_end=None, delta_t=None, der=0):
        """Python generator that yields locations along the trajectory.

        Parameters
        ----------
        t_start : float
            Starting time of the trajectory, defaults to the earliest
            time in :attr:`points`.
        t_end : float
            Ending time of the trajectory, defaults to the latest
            time in :attr:`points`.
        delta_t : float
            Time interval between yielded trajectory points, defaults to earliest
            time in :attr:`points`.

        Returns
        -------
        (x, y, z) : tuples of floats
            Positions at the desired times are yielded.

        Examples
        --------
        x.traj(0.1)
            Yields the position every 0.1 s within the
            given :attr:`interval`.
        x.traj(2.5, 4.5, 0.1)
            Yields the position every 0.1 s between 2.5 s and 4.5 s.
        x.traj(0.1, der=1)
            Yields the 1st derivative of the spline (= velocity vector) every 0.1 s
            within the given :attr:`interval`.

        """
        if not delta_t:
            delta_t = t_start
            t_start, t_end = self.interval
        if not t_end:
            t_end = self.interval[1]
        # all locations are fetched in one go because that is much faster further improvement could
        # be possible if interpolated locations are fetched in blocks
        yield from zip(*self.location(arange(t_start, t_end, delta_t), der))



# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------
"""Implements signal generators for the simulation of acoustic sources.

.. autosummary::
    :toctree: generated/

    SignalGenerator
    PeriodicSignalGenerator
    NoiseGenerator
    WNoiseGenerator
    PNoiseGenerator
    FiltWNoiseGenerator
    SineGenerator
    GenericSignalGenerator

"""

# imports from other packages
from abc import abstractmethod
from warnings import warn

from numpy import arange, array, log, pi, repeat, sin, sqrt, tile, zeros
from numpy.random import RandomState
from scipy.signal import resample, sosfilt, tf2sos
from traits.api import (
    ABCHasStrictTraits,
    Bool,
    CArray,
    CInt,
    Delegate,
    Float,
    Instance,
    Int,
    Property,
    cached_property,
)

# acoular imports
from .base import SamplesGenerator
from .deprecation import deprecated_alias
from .internal import digest


@deprecated_alias({'numsamples': 'num_samples'})
class SignalGenerator(ABCHasStrictTraits):
    """Virtual base class for a simple one-channel signal generator.

    Defines the common interface for all SignalGenerator classes. This class
    may be used as a base for specialized SignalGenerator implementations. It
    should not be used directly as it contains no real functionality.
    """

    #: Sampling frequency of the signal.
    sample_freq = Float(1.0, desc='sampling frequency')

    #: Number of samples to generate.
    num_samples = CInt

    # internal identifier
    digest = Property(depends_on=['sample_freq', 'num_samples'])

    @abstractmethod
    def _get_digest(self):
        """Returns the internal identifier."""

    @abstractmethod
    def signal(self):
        """Deliver the signal."""

    def usignal(self, factor):
        """Delivers the signal resampled with a multiple of the sampling freq.

        Uses fourier transform method for resampling (from scipy.signal).

        Parameters
        ----------
        factor : integer
            The factor defines how many times the new sampling frequency is
            larger than :attr:`sample_freq`.

        Returns
        -------
        array of floats
            The resulting signal of length `factor` * :attr:`num_samples`.
        """
        return resample(self.signal(), factor * self.num_samples)


class PeriodicSignalGenerator(SignalGenerator):
    """
    Abstract base class for periodic signal generators.

    Defines the common interface for all :class:`SignalGenerator`-derived classes with periodic
    signals. This class may be used as a base for class handling periodic signals that can be
    characterized by their frequency, phase and amplitude. It should not be used directly as it
    contains no real functionality.
    """

    #: Frequency of the signal, float, defaults to 1000.0.
    freq = Float(1000.0, desc='Frequency')

    #: Phase of the signal (in radians), float, defaults to 0.0.
    phase = Float(0.0, desc='Phase')

    #: Amplitude of the signal. Defaults to 1.0.
    amplitude = Float(1.0)

    # internal identifier
    digest = Property(depends_on=['amplitude', 'num_samples', 'sample_freq', 'freq', 'phase'])

    @abstractmethod
    def _get_digest(self):
        """Returns the internal identifier."""

    @abstractmethod
    def signal(self):
        """Deliver the signal."""


class NoiseGenerator(SignalGenerator):
    """Abstract base class for noise signal generators.

    Defines the common interface for all :class:`SignalGenerator` classes with noise signals. This
    class may be used as a base for class handling noise signals that can be characterized by their
    RMS amplitude. It should not be used directly as it contains no real functionality.
    """

    #: RMS amplitude of the signal.
    rms = Float(1.0, desc='rms amplitude')

    #: Seed for random number generator, defaults to 0.
    #: This parameter should be set differently for different instances
    #: to guarantee statistically independent (non-correlated) outputs.
    seed = Int(0, desc='random seed value')

    # internal identifier
    digest = Property(depends_on=['rms', 'seed', 'sample_freq', 'num_samples'])

    @abstractmethod
    def _get_digest(self):
        """Returns the internal identifier."""

    @abstractmethod
    def signal(self):
        """Deliver the signal."""


class WNoiseGenerator(NoiseGenerator):
    """White noise signal generator."""

    # internal identifier
    digest = Property(depends_on=['rms', 'seed', 'sample_freq', 'num_samples'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    def signal(self):
        """Deliver the signal.

        Returns
        -------
        Array of floats
            The resulting signal as an array of length :attr:`~SignalGenerator.num_samples`.
        """
        rnd_gen = RandomState(self.seed)
        return self.rms * rnd_gen.standard_normal(self.num_samples)


class PNoiseGenerator(NoiseGenerator):
    """Pink noise signal generator.

    Simulation of pink noise is based on the Voss-McCartney algorithm.
    Ref.:

      * S.J. Orfanidis: Signal Processing (2010), pp. 729-733
      * online discussion: http://www.firstpr.com.au/dsp/pink-noise/

    The idea is to iteratively add larger-wavelength noise to get 1/f
    characteristic.
    """

    #: "Octave depth" -- higher values for 1/f spectrum at low frequencies,
    #: but longer calculation, defaults to 16.
    depth = Int(16, desc='octave depth')

    # internal identifier
    digest = Property(depends_on=['rms', 'seed', 'sample_freq', 'num_samples', 'depth'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    def signal(self):
        nums = self.num_samples
        depth = self.depth
        # maximum depth depending on number of samples
        max_depth = int(log(nums) / log(2))

        if depth > max_depth:
            depth = max_depth
            print(f'Pink noise filter depth set to maximum possible value of {max_depth:d}.')

        rnd_gen = RandomState(self.seed)
        s = rnd_gen.standard_normal(nums)
        for _ in range(depth):
            ind = 2**_ - 1
            lind = nums - ind
            dind = 2 ** (_ + 1)
            s[ind:] += repeat(rnd_gen.standard_normal(nums // dind + 1), dind)[:lind]
        # divide by sqrt(depth+1.5) to get same overall level as white noise
        return self.rms / sqrt(depth + 1.5) * s


class FiltWNoiseGenerator(WNoiseGenerator):
    """Filtered white noise signal following an autoregressive (AR), moving-average
    (MA) or autoregressive moving-average (ARMA) process.

    The desired frequency response of the filter can be defined by specifying
    the filter coefficients :attr:`ar` and :attr:`ma`.
    The RMS value specified via the :attr:`rms` attribute belongs to the white noise
    signal and differs from the RMS value of the filtered signal.
    For numerical stability at high orders, the filter is a combination of second order
    sections (sos).
    """

    ar = CArray(value=array([]), dtype=float, desc='autoregressive coefficients (coefficients of the denominator)')

    ma = CArray(value=array([]), dtype=float, desc='moving-average coefficients (coefficients of the numerator)')

    # internal identifier
    digest = Property(depends_on=['rms', 'seed', 'sample_freq', 'num_samples', 'ar', 'ma'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    def handle_empty_coefficients(self, coefficients):
        if coefficients.size == 0:
            return array([1.0])
        return coefficients

    def signal(self):
        """Deliver the signal.

        Returns
        -------
        Array of floats
            The resulting signal as an array of length :attr:`~SignalGenerator.num_samples`.
        """
        rnd_gen = RandomState(self.seed)
        ma = self.handle_empty_coefficients(self.ma)
        ar = self.handle_empty_coefficients(self.ar)
        sos = tf2sos(ma, ar)
        ntaps = ma.shape[0]
        sdelay = round(0.5 * (ntaps - 1))
        wnoise = self.rms * rnd_gen.standard_normal(
            self.num_samples + sdelay,
        )  # create longer signal to compensate delay
        return sosfilt(sos, x=wnoise)[sdelay:]


class SineGenerator(PeriodicSignalGenerator):
    """Sine signal generator with adjustable amplitude, frequency and phase."""

    # internal identifier
    digest = Property(depends_on=['num_samples', 'sample_freq', 'amplitude', 'freq', 'phase'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    def signal(self):
        """Deliver the signal.

        Returns
        -------
        array of floats
            The resulting signal as an array of length :attr:`~SignalGenerator.num_samples`.
        """
        t = arange(self.num_samples, dtype=float) / self.sample_freq
        return self.amplitude * sin(2 * pi * self.freq * t + self.phase)


@deprecated_alias({'rms': 'amplitude'})
class GenericSignalGenerator(SignalGenerator):
    """Generate signal from output of :class:`~acoular.base.SamplesGenerator` object.

    This class can be used to inject arbitrary signals into Acoular processing
    chains. For example, it can be used to read signals from a HDF5 file or create any signal
    by using the :class:`acoular.sources.TimeSamples` class.

    Example
    -------
    >>> import numpy as np
    >>> from acoular import TimeSamples, GenericSignalGenerator
    >>> data = np.random.rand(1000, 1)
    >>> ts = TimeSamples(data=data, sample_freq=51200)
    >>> sig = GenericSignalGenerator(source=ts)
    """

    #: Data source; :class:`~acoular.base.SamplesGenerator` or derived object.
    source = Instance(SamplesGenerator)

    #: Amplitude of the signal. Defaults to 1.0.
    amplitude = Float(1.0)

    #: Sampling frequency of output signal, as given by :attr:`source`.
    sample_freq = Delegate('source')

    _num_samples = CInt(0)

    #: Number of samples to generate. Is set to source.num_samples by default.
    num_samples = Property()

    def _get_num_samples(self):
        if self._num_samples:
            return self._num_samples
        return self.source.num_samples

    def _set_num_samples(self, num_samples):
        self._num_samples = num_samples

    #: Boolean flag, if 'True' (default), signal track is repeated if requested
    #: :attr:`num_samples` is higher than available sample number
    loop_signal = Bool(True)

    # internal identifier
    digest = Property(
        depends_on=['source.digest', 'loop_signal', 'num_samples', 'amplitude'],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    def signal(self):
        """Deliver the signal.

        Returns
        -------
        array of floats
            The resulting signal as an array of length :attr:`~GenericSignalGenerator.num_samples`.

        """
        block = 1024
        if self.source.num_channels > 1:
            warn(
                'Signal source has more than one channel. Only channel 0 will be used for signal.',
                Warning,
                stacklevel=2,
            )
        nums = self.num_samples
        track = zeros(nums)

        # iterate through source generator to fill signal track
        for i, temp in enumerate(self.source.result(block)):
            start = block * i
            stop = start + len(temp[:, 0])
            if nums > stop:
                track[start:stop] = temp[:, 0]
            else:  # exit loop preliminarily if wanted signal samples are reached
                track[start:nums] = temp[: nums - start, 0]
                break

        # if the signal should be repeated after finishing and there are still samples open
        if self.loop_signal and (nums > stop):
            # fill up empty track with as many full source signals as possible
            nloops = nums // stop
            if nloops > 1:
                track[stop : stop * nloops] = tile(track[:stop], nloops - 1)
            # fill up remaining empty track
            res = nums % stop  # last part of unfinished loop
            if res > 0:
                track[stop * nloops :] = track[:res]
        return self.amplitude * track



# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------
"""General purpose blockwise processing methods independent of the domain (time or frequency).

.. autosummary::
    :toctree: generated/

    Average
    Cache
    SampleSplitter
    TimeAverage
    TimeCache
    SamplesBuffer
"""

import threading
from collections import deque
from inspect import currentframe
from warnings import warn

import numpy as np
from traits.api import Any, Array, Bool, Dict, Enum, Instance, Int, Property, Union, cached_property, on_trait_change

# acoular imports
from .base import Generator, InOut
from .configuration import config
from .deprecation import deprecated_alias
from .h5cache import H5cache
from .h5files import H5CacheFileBase
from .internal import digest
from .tools.utils import find_basename


class LockedGenerator:
    """Creates a Thread Safe Iterator.
    Takes an iterator/generator and makes it thread-safe by
    serializing call to the `next` method of given iterator/generator.
    """

    def __init__(self, it):
        self.it = it
        self.lock = threading.Lock()

    def __next__(self):
        with self.lock:
            return self.it.__next__()


@deprecated_alias({'naverage': 'num_per_average', 'numsamples': 'num_samples'}, read_only=['numsamples'])
class Average(InOut):
    """Calculates the average across consecutive time samples or frequency snapshots.

    The average operation is performed differently depending on the source type.
    If the source is a time domain source
    (e.g. derived from :class:`~acoular.base.SamplesGenerator`), the average is
    calculated over a certain number of time samples given by :attr:`num_per_average`.
    If the source is a frequency domain source (e.g. derived from
    :class:`~acoular.base.SpectraGenerator`), the average is calculated over a certain
    number of snapshots given by :attr:`num_per_average`.

    Examples
    --------
        For estimate the RMS of a white noise (time-domain) signal, the average of the squared
        signal can be calculated:

        >>> import acoular as ac
        >>> import numpy as np
        >>>
        >>> signal = ac.WNoiseGenerator(sample_freq=51200, num_samples=51200, rms=2.0).signal()
        >>> ts = ac.TimeSamples(data=signal[:, np.newaxis], sample_freq=51200)
        >>> tp = ac.TimePower(source=ts)
        >>> avg = ac.Average(source=tp, num_per_average=512)
        >>> mean_squared_value = next(avg.result(num=1))
        >>> rms = np.sqrt(mean_squared_value)[0, 0]
        >>> print(rms)
        1.9985200025816718

        Here, each evaluation of the generator created by the :meth:`result` method of the
        :class:`Average` object via the :meth:`next` function returns :code:`num=1` average across a
        snapshot of 512 samples.

        If the source is a frequency domain source, the average is calculated over a certain number
        of snapshots, defined by :attr:`num_per_average`.

        >>> fft = ac.RFFT(source=ts, block_size=64)
        >>> ps = ac.AutoPowerSpectra(source=fft)
        >>> avg = ac.Average(source=ps, num_per_average=16)
        >>> mean_power = next(avg.result(num=1))
        >>> print(np.sqrt(mean_power.sum()))
        2.0024960894399295

        Here, the generator created by the :meth:`result` method of the :class:`Average` object
        returns the average across 16 snapshots in the frequency domain.

    """

    #: Number of samples (time domain source) or snapshots (frequency domain source)
    #: to average over, defaults to 64.
    num_per_average = Int(64, desc='number of samples/snapshots to average over')

    #: Sampling frequency of the output signal, is set automatically.
    sample_freq = Property(depends_on=['source.sample_freq', 'num_per_average'])

    #: Number of samples (time domain) or snapshots (frequency domain) of the output signal.
    #: Is set automatically.
    num_samples = Property(depends_on=['source.num_samples', 'num_per_average'])

    # internal identifier
    digest = Property(depends_on=['source.digest', 'num_per_average'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_sample_freq(self):
        if self.source:
            return 1.0 * self.source.sample_freq / self.num_per_average
        return None

    @cached_property
    def _get_num_samples(self):
        if self.source:
            return self.source.num_samples / self.num_per_average
        return None

    def result(self, num):
        """Python generator that yields the output block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Average of the output of source.
            Yields samples in blocks of shape (num, num_channels).
            The last block may be shorter than num.

        """
        nav = self.num_per_average
        for temp in self.source.result(num * nav):
            ns, nc = temp.shape
            nso = int(ns / nav)
            if nso > 0:
                yield temp[: nso * nav].reshape((nso, -1, nc)).mean(axis=1)


class Cache(InOut):
    """Caches source output in cache file.

    This class is used to cache the output of a :class:`acoular.base.Generator` derived source
    object in a cache file to circumvent time-consuming re-calculation.
    The cache file is created in the Acoular cache directory.

    Examples
    --------
    >>> import acoular as ac
    >>> import numpy as np
    >>>
    >>> ac.config.h5library = 'tables'
    >>> data = np.random.rand(1024, 1)
    >>> ts = ac.TimeSamples(data=data, sample_freq=51200)
    >>> fft = ac.RFFT(source=ts, block_size=1024)
    >>> cache = ac.Cache(source=fft)  # cache the output of the FFT in cache file
    >>> for block in cache.result(num=1):  # read the cached data block-wise
    ...     print(block.shape)
    [('void_cache.h5', 1)]
    (1, 513)

    The caching behaviour can be controlled by the :class:`~acoular.configuration.Config` instance
    via the :attr:`~acoular.configuration.Config.global_caching` attribute.
    To turn off caching, set :attr:`~acoular.configuration.Config.global_caching` to 'none' before
    running the code. The cache file directory can be obtained (and set) via the
    :attr:`~acoular.configuration.Config.cache_dir`

    >>> ac.config.global_caching = 'none'

    """

    # basename for cache
    basename = Property(depends_on=['digest'])

    # hdf5 cache file
    h5f = Instance(H5CacheFileBase, transient=True)

    # internal identifier
    digest = Property(depends_on=['source.digest'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_basename(self):
        return find_basename(self.source)

    def _pass_data(self, num):
        yield from self.source.result(num)

    def _write_data_to_cache(self, num):
        nodename = 'tc_' + self.digest
        for i, data in enumerate(self.source.result(num)):
            if i == 0:
                self.h5f.create_extendable_array(nodename, (0, data.shape[1]), data.dtype.name)
                ac = self.h5f.get_data_by_reference(nodename)
                self.h5f.set_node_attribute(ac, 'sample_freq', self.sample_freq)
                self.h5f.set_node_attribute(ac, 'complete', False)
            self.h5f.append_data(ac, data)
            self.h5f.flush()
            yield data
        self.h5f.set_node_attribute(ac, 'complete', True)

    def _get_data_from_cache(self, num):
        nodename = 'tc_' + self.digest
        ac = self.h5f.get_data_by_reference(nodename)
        i = 0
        while i < ac.shape[0]:
            yield ac[i : i + num]
            i += num

    def _get_data_from_incomplete_cache(self, num):
        nodename = 'tc_' + self.digest
        ac = self.h5f.get_data_by_reference(nodename)
        i = 0
        nblocks = 0
        while i + num <= ac.shape[0]:
            yield ac[i : i + num]
            nblocks += 1
            i += num
        self.h5f.remove_data(nodename)
        for j, data in enumerate(self.source.result(num)):
            if j == 0:
                self.h5f.create_extendable_array(nodename, (0, data.shape[1]), data.dtype.name)
                ac = self.h5f.get_data_by_reference(nodename)
                self.h5f.set_node_attribute(ac, 'sample_freq', self.sample_freq)
                self.h5f.set_node_attribute(ac, 'complete', False)
            self.h5f.append_data(ac, data)
            if j >= nblocks:
                self.h5f.flush()
                yield data
        self.h5f.set_node_attribute(ac, 'complete', True)

    # result generator: delivers input, possibly from cache
    def result(self, num):
        """Python generator that yields the output from cache block-wise.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Samples in blocks of shape (num, num_channels).
            The last block may be shorter than num.
            Echos the source output, but reads it from cache
            when available and prevents unnecessary recalculation.

        """
        if config.global_caching == 'none':
            generator = self._pass_data
        else:
            nodename = 'tc_' + self.digest
            H5cache.get_cache_file(self, self.basename)
            if not self.h5f:
                generator = self._pass_data
            elif self.h5f.is_cached(nodename):
                generator = self._get_data_from_cache
                if config.global_caching == 'overwrite':
                    self.h5f.remove_data(nodename)
                    generator = self._write_data_to_cache
                elif not self.h5f.get_data_by_reference(nodename).attrs.__contains__('complete'):
                    if config.global_caching == 'readonly':
                        generator = self._pass_data
                    else:
                        generator = self._get_data_from_incomplete_cache
                elif not self.h5f.get_data_by_reference(nodename).attrs['complete']:
                    if config.global_caching == 'readonly':
                        warn(
                            f"Cache file is incomplete for nodename {nodename}. With config.global_caching='readonly', \
                            the cache file will not be used!",
                            Warning,
                            stacklevel=1,
                        )
                        generator = self._pass_data
                    else:
                        generator = self._get_data_from_incomplete_cache
            elif not self.h5f.is_cached(nodename):
                generator = self._write_data_to_cache
                if config.global_caching == 'readonly':
                    generator = self._pass_data
        yield from generator(num)


class SampleSplitter(InOut):
    """
    Distributes data from a source to several following objects in a block-wise manner.

    The `SampleSplitter` class is designed to take data from a single
    :class:`~acoular.base.Generator` derived source object and distribute it to multiple
    :class:`~acoular.base.Generator` derived objects. For each object, the :class:`SampleSplitter`
    holds a virtual block buffer from which the subsequently connected objects receive data in a
    first-in-first-out (FIFO) manner. This allows for efficient data handling and processing in
    parallel.

    Examples
    --------
    Consider a time domain source signal stream from which the FFT spectra and the signal power
    are calculated block-wise and in parallel by using the :class:`~acoular.fprocess.RFFT` as well
    as the :class:`~acoular.tprocess.TimePower` and :class:`~acoular.process.Average`
    objects. The `SampleSplitter` object is used to distribute the incoming blocks of data to the
    `RFFT` and `TimePower` object buffers whenever one of these objects calls the :meth:`result`
    generator.
    For the `TimePower` object, the buffer size is set to 10 blocks. If the buffer is full, an error
    is raised since the buffer overflow treatment is set to 'error'. For the `RFFT` object, the
    block buffer size is set to 1 block, and the buffer overflow treatment is set to 'none'. This
    is done to reduce latency in the FFT calculation, as the FFT calculation may take longer than
    the signal power calculation. If new data is available and the block buffer for the `RFFT`
    object is full, the `SampleSplitter` will drop the oldest block of data in the buffer. Thus, the
    `RFFT` object will always receive the most recent block of data.

    >>> import acoular as ac
    >>> import numpy as np
    >>>
    >>> # create a time domain signal source
    >>> ts = ac.TimeSamples(data=np.random.rand(1024, 1), sample_freq=51200)
    >>>
    >>> # create the sample splitter object
    >>> ss = ac.SampleSplitter(source=ts)
    >>>
    >>> # create the FFT spectra and further objects that receive the data
    >>> fft = ac.RFFT(source=ss, block_size=64)
    >>> pow = ac.TimePower(source=ss)
    >>> avg = ac.Average(source=pow, num_per_average=64)
    >>>
    >>> # register the subsequent processing block objects at the sample splitter
    >>> ss.register_object(fft, buffer_size=1, buffer_overflow_treatment='none')
    >>> ss.register_object(pow, buffer_size=10, buffer_overflow_treatment='error')

    After object registration, the `SampleSplitter` object is ready to distribute the data to the
    object buffers. The block buffers can be accessed via the `block_buffer` attribute of the
    `SampleSplitter` object.

    >>> ss.block_buffer.values()
    dict_values([deque([], maxlen=1), deque([], maxlen=10)])

    Calling the result method of the FFT object will start the data collection and distribution
    process.

    >>> generator = fft.result(num=1)
    >>> fft_res = next(generator)

    Although we haven't called the result method of the signal power object, one data block is
    already available in the buffer.

    >>> print(len(ss.block_buffer[pow]))
    1

    To remove registered objects from the `SampleSplitter`, use the :meth:`remove_object` method.

    >>> ss.remove_object(pow)
    >>> print(len(ss.block_buffer))
    1
    """

    #: dictionary with block buffers (dict values) of registered objects (dict
    #: keys).
    block_buffer = Dict(key_trait=Instance(Generator))

    #: max elements/blocks in block buffers.
    #: Can be set individually for each registered object.
    #: Default is 100 blocks for each registered object.
    buffer_size = Union(
        Int,
        Dict(key_trait=Instance(Generator), value_trait=Int),
        default_value=100,
    )

    #: defines behaviour in case of block_buffer overflow. Can be set individually
    #: for each registered object.
    #:
    #: * 'error': an IOError is thrown by the class
    #: * 'warning': a warning is displayed. Possibly leads to lost blocks of data
    #: * 'none': nothing happens. Possibly leads to lost blocks of data
    buffer_overflow_treatment = Dict(
        key_trait=Instance(Generator),
        value_trait=Enum('error', 'warning', 'none'),
        desc='defines buffer overflow behaviour.',
    )

    # shadow trait to monitor if source deliver samples or is empty
    _source_generator_exist = Bool(False)

    # shadow trait to monitor if buffer of objects with overflow treatment = 'error'
    # or warning is overfilled. Error will be raised in all threads.
    _buffer_overflow = Bool(False)

    # Helper Trait holds source generator
    _source_generator = Instance(LockedGenerator)

    def _create_block_buffer(self, obj, buffer_size=None):
        if buffer_size is None:
            buffer_size = self.buffer_size if isinstance(self.buffer_size, int) else self.buffer_size[obj]
        self.block_buffer[obj] = deque([], maxlen=buffer_size)

    def _create_buffer_overflow_treatment(self, obj, buffer_overflow_treatment=None):
        if buffer_overflow_treatment is None:
            buffer_overflow_treatment = 'error'
        self.buffer_overflow_treatment[obj] = buffer_overflow_treatment

    def _clear_block_buffer(self, obj):
        self.block_buffer[obj].clear()

    def _remove_block_buffer(self, obj):
        del self.block_buffer[obj]

    def _remove_buffer_overflow_treatment(self, obj):
        del self.buffer_overflow_treatment[obj]

    def _assert_obj_registered(self, obj):
        if obj not in self.block_buffer:
            msg = f'calling object {obj} is not registered.'
            raise OSError(msg)

    def _get_objs_to_inspect(self):
        return [obj for obj in self.buffer_overflow_treatment if self.buffer_overflow_treatment[obj] != 'none']

    def _inspect_buffer_levels(self, inspect_objs):
        for obj in inspect_objs:
            if len(self.block_buffer[obj]) == self.buffer_size:
                if self.buffer_overflow_treatment[obj] == 'error':
                    self._buffer_overflow = True
                elif self.buffer_overflow_treatment[obj] == 'warning':
                    warn(f'overfilled buffer for object: {obj} data will get lost', UserWarning, stacklevel=1)

    def _create_source_generator(self, num):
        for obj in self.block_buffer:
            self._clear_block_buffer(obj)
        self._buffer_overflow = False  # reset overflow bool
        self._source_generator = LockedGenerator(self.source.result(num))
        self._source_generator_exist = True  # indicates full generator

    def _fill_block_buffers(self):
        next_block = next(self._source_generator)
        [self.block_buffer[obj].appendleft(next_block) for obj in self.block_buffer]

    @on_trait_change('buffer_size')
    def _change_buffer_size(self):  #
        for obj in self.block_buffer:
            self._remove_block_buffer(obj)
            self._create_block_buffer(obj)

    def register_object(self, *objects_to_register, buffer_size=None, buffer_overflow_treatment=None):
        """Register one or multiple :class:`~acoular.base.Generator` objects to the SampleSplitter.

        Creates a block buffer for each object and sets the buffer size and buffer
        overflow treatment.

        Parameters
        ----------
        objects_to_register : Generator
            One or multiple :class:`~acoular.base.Generator` derived objects to be registered.
        buffer_size : int, optional
            Maximum number of elements/blocks in block buffer. If not set, the default buffer size
            of 100 blocks is used.
        buffer_overflow_treatment : str, optional
            Defines the behaviour in case of reaching the buffer size.
            Can be set individually for each object. Possible values are 'error', 'warning', and
            'none'. If not set, the default value is 'error'.
        """
        for obj in objects_to_register:
            if obj not in self.block_buffer:
                self._create_block_buffer(obj, buffer_size)
                self._create_buffer_overflow_treatment(obj, buffer_overflow_treatment)
            else:
                msg = f'object {obj} is already registered.'
                raise OSError(msg)

    def remove_object(self, *objects_to_remove):
        """Function that can be used to remove registered objects.

        If no objects are given, all registered objects are removed.

        Parameters
        ----------
        objects_to_remove : list
            One or multiple :class:`~acoular.base.Generator` derived objects to be removed.
            If not set, all registered objects are removed.
        """
        if not objects_to_remove:
            objects_to_remove = list(self.block_buffer.keys())
        for obj in objects_to_remove:
            self._remove_block_buffer(obj)
            self._remove_buffer_overflow_treatment(obj)

    def result(self, num):
        """Python generator that yields the output block-wise from block-buffer.

        Parameters
        ----------
        num : integer
            This parameter defines the size of the blocks to be yielded
            (i.e. the number of samples per block).

        Returns
        -------
        Samples in blocks of shape (num, num_channels).
            Delivers a block of samples to the calling object.
            The last block may be shorter than num.

        """
        calling_obj = currentframe().f_back.f_locals['self']
        self._assert_obj_registered(calling_obj)
        objs_to_inspect = self._get_objs_to_inspect()

        if not self._source_generator_exist:
            self._create_source_generator(num)

        while not self._buffer_overflow:
            if self.block_buffer[calling_obj]:
                yield self.block_buffer[calling_obj].pop()
            else:
                self._inspect_buffer_levels(objs_to_inspect)
                try:
                    self._fill_block_buffers()
                except StopIteration:
                    self._source_generator_exist = False
                    return
        else:
            msg = 'Maximum size of block buffer is reached!'
            raise OSError(msg)


class TimeAverage(Average):
    """Calculates average of the signal (Alias for :class:`acoular.process.Average`).

    .. deprecated:: 24.10
        Using :class:`~acoular.process.TimeAverage` is deprecated and will be removed in Acoular
        version 25.07. Use :class:`~acoular.process.Average` instead.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warn(
            'Using TimeAverage is deprecated and will be removed in Acoular version 25.07. Use Average instead.',
            DeprecationWarning,
            stacklevel=2,
        )


class TimeCache(Cache):
    """Caches source signals in cache file (Alias for :class:`acoular.process.Cache`).

    .. deprecated:: 24.10
        Using :class:`~acoular.process.TimeCache` is deprecated and will be removed in Acoular
        version 25.07. Use :class:`~acoular.process.Cache` instead.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warn(
            'Using TimeCache is deprecated and will be removed in Acoular version 25.07. Use Cache instead.',
            DeprecationWarning,
            stacklevel=2,
        )


class SamplesBuffer(InOut):
    """Handles buffering of samples from a source.

    This class is used to buffer samples from a source and provide them in blocks
    of a specified size. There are several usecases for this class, as demonstrated in
    the following.

    Examples
    --------
    Let us assume we want to draw blocks of 16 samples from our source, but we want to make sure
    that we always have twice the number of samples in the buffer. We can achieve this simple
    behaviour by using the following code:

    >>> import acoular as ac
    >>> import numpy as np
    >>> # create a white noise source with 512 samples
    >>> source = ac.TimeSamples(
    ...     data=ac.WNoiseGenerator(
    ...         sample_freq=64,
    ...         num_samples=512,
    ...     ).signal()[:, np.newaxis],
    ...     sample_freq=64,
    ... )
    >>> # create a buffer with a size of 32 samples
    >>> buffer = ac.process.SamplesBuffer(source=source, length=32)
    >>> # get the first block of 16 samples
    >>> block = next(buffer.result(num=16))
    >>> np.testing.assert_array_equal(block, source.data[:16])

    Here, on the first call to the result method, the buffer will fill up by collecting blocks with
    same size from the source. The buffer will then return the first block of 16 samples. On the
    next call to the result method, the buffer will be filled again and returns the next block of 16
    samples.

    In some cases, we might want to draw a different number of samples from the source than we want
    to return. This can be achieved by setting the `source_num` trait of the buffer. A special case
    is the return of a variable number of samples. This is the case, for example, in the class
    :class:`~acoular.tbeamform.BeamformerTimeTraj`, in which a different number of time samples is
    required from the buffer for further delay-and-sum processing depending on the expected delay,
    which can be vary for moving sources. At the same time, however, only 'num' samples should be
    written to and removed from the buffer. This behavior can be achieved by setting the
    `shift_index_by` trait to 'num' and by setting the `result_num` trait to the number of samples
    that should be returned by the result function.

    >>> buffer = ac.process.SamplesBuffer(source=source, length=32, result_num=20, shift_index_by='num')
    >>> block_sizes = []
    >>> block_sizes.append(
    ...     next(buffer.result(num=16)).shape[0]
    ... )  # this time, the buffer will return 20 samples, but the buffer will only forget the first 16 samples
    >>> buffer.result_num = 24
    >>> block_sizes.append(
    ...     next(buffer.result(num=16)).shape[0]
    ... )  # this time, the buffer will return 24 samples, but the buffer will only forget the first 16 samples
    >>> np.testing.assert_array_equal(block_sizes, [20, 24])
    """  # noqa: W505

    #: number of samples that fit in the buffer
    length = Int(desc='number of samples that fit in the buffer')

    #: number of samples per block to obtain from the source. If 'None', use 'num' argument of
    #: result method
    source_num = Union(
        None,
        Int(),
        default_value=None,
        desc='number of samples to return from the source. If "None", use "num" argument of result method',
    )

    #: number of samples to return from the buffer. If 'None', use 'num' argument of result method
    result_num = Union(
        None,
        Int(),
        default_value=None,
        desc="number of samples to return from the buffer. If 'None', use 'num' argument of result method",
    )

    #: index shift value for the buffer. If "result_num", buffer will return and forget 'result_num'
    #: samples. If "num", buffer will return 'result_num' samples but will forget 'num' samples
    shift_index_by = Enum(
        ('result_num', 'num'),
        desc=(
            'index shift value for the buffer. If "result_num", use "result_num" trait.'
            ' If "num", use "num" argument of result method'
        ),
    )

    #: current filling level of buffer
    level = Property(desc='current filling level of buffer')

    #: data type of the buffer elements
    dtype = Any(desc='data type of the buffer')

    # flag to indicate that the source is empty, for internal use
    _empty_source = Bool(False, desc='flag to indicate that the source is empty')

    # the buffer for processing
    _buffer = Array(shape=(None, None), desc='buffer for block processing')

    # current index in buffer
    _index = Int(desc='current index in buffer')

    def _get_level(self):
        return self._buffer.shape[0] - self._index

    def _create_new_buffer(self):
        self._buffer = np.zeros((self.length, self.num_channels), dtype=self.dtype)
        self._index = self.length
        self._empty_source = False

    def _write_to_buffer(self, data):
        ns = data.shape[0]
        self._buffer[0 : (self.length - ns)] = self._buffer[-(self.length - ns) :]
        self._buffer[-ns:, :] = data.astype(self.dtype)
        self._index -= ns

    def increase_buffer(self, num):
        """Increase the buffer by 'num' samples.

        Returns
        -------
        None
        """
        ar = np.zeros((num, self.num_channels), dtype=self._buffer.dtype)
        self._buffer = np.concatenate((ar, self._buffer), axis=0)
        self._index += num
        self.length += num

    def read_from_buffer(self, num):
        """Read samples from the buffer.

        Parameters
        ----------
        num : int
            number of samples to read from the buffer.

        Returns
        -------
        numpy.ndarray
            block of samples from the buffer

        """
        rnum = num if self.result_num is None else self.result_num
        rnum = rnum if self.level >= rnum else self.level
        data = self._buffer[self._index : self._index + rnum]
        if self.shift_index_by == 'result_num':
            self._index += rnum
        else:
            self._index += num
        return data

    def fill_buffer(self, snum):
        """Fill the buffer with samples from the source.

        Parameters
        ----------
        snum : int
            number of samples to return from the source.

        Yields
        ------
        None
        """
        source_generator = self.source.result(snum)
        while not self._empty_source:
            while self._index >= snum:
                if self.result_num is not None:
                    while self.result_num > self.length:
                        self.increase_buffer(snum)
                try:
                    self._write_to_buffer(next(source_generator))
                except StopIteration:
                    self._empty_source = True
                    break
            yield

    def result(self, num):
        """Return blocks of samples from the buffer.

        Parameters
        ----------
        num : int
            number of samples to return.

        Yields
        ------
        numpy.ndarray
            block of samples from the buffer
        """
        self._create_new_buffer()
        snum = num
        if self.source_num is not None:
            snum = self.source_num
        for _ in self.fill_buffer(snum):
            if self.level > 0:
                yield self.read_from_buffer(num)
            else:
                break
        while self.level > 0:
            yield self.read_from_buffer(num)



# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------
"""Implements support for two- and threedimensional grids.

.. autosummary::
    :toctree: generated/

    Grid
    RectGrid
    RectGrid3D
    ImportGrid
    LineGrid
    MergeGrid
    Sector
    SingleSector
    RectSector
    RectSector3D
    CircSector
    PolySector
    ConvexSector
    MultiSector

"""

# imports from other packages
import xml.dom.minidom
from abc import abstractmethod

from numpy import (
    absolute,
    append,
    arange,
    argmin,
    array,
    asarray,
    concatenate,
    copysign,
    fabs,
    inf,
    isscalar,
    ma,
    mgrid,
    newaxis,
    ones,
    ones_like,
    s_,
    sum,  # noqa: A004
    tile,
    unique,
    where,
    zeros,
)
from scipy.linalg import norm

# from matplotlib.path import Path
from scipy.spatial import Delaunay
from traits.api import (
    ABCHasStrictTraits,
    Bool,
    CArray,
    File,
    Float,
    Instance,
    Int,
    List,
    Property,
    Str,
    Tuple,
    Union,
    cached_property,
    observe,
    on_trait_change,
    property_depends_on,
)
from traits.trait_errors import TraitError

# acoular imports
from .deprecation import deprecated_alias
from .internal import digest, ldigest


def in_hull(p, hull, border=True, tol=0):
    """Test if points in `p` are in `hull`
    `p` should be a `NxK` coordinates of `N` points in `K` dimensions
    `hull` is either a scipy.spatial.Delaunay object or the `MxK` array of the
    coordinates of `M` points in `K`dimensions for which Delaunay triangulation
    will be computed.
    """
    if not isinstance(hull, Delaunay):
        hull = Delaunay(hull)

    if border:
        return hull.find_simplex(p, tol=tol) >= 0
    return hull.find_simplex(p, tol=tol) > 0


def _det(xvert, yvert):
    """Compute twice the area of the triangle defined by points with using
    determinant formula.

    Input parameters:

    xvert -- A vector of nodal x-coords (array-like).
    yvert -- A vector of nodal y-coords (array-like).

    Output parameters:
    Twice the area of the triangle defined by the points.

    Notes
    -----
    _det is positive if points define polygon in anticlockwise order.
    _det is negative if points define polygon in clockwise order.
    _det is zero if at least two of the points are concident or if
        all points are collinear.

    """
    xvert = asarray(xvert, dtype=float)
    yvert = asarray(yvert, dtype=float)
    x_prev = concatenate(([xvert[-1]], xvert[:-1]))
    y_prev = concatenate(([yvert[-1]], yvert[:-1]))
    return sum(yvert * x_prev - xvert * y_prev, axis=0)


class Polygon:
    """Polygon object.
    Input parameters:
    x -- A sequence of nodal x-coords.
    y -- A sequence of nodal y-coords.
    """

    def __init__(self, x, y):
        if len(x) != len(y):
            msg = 'x and y must be equally sized.'
            raise IndexError(msg)
        self.x = asarray(x, dtype=float)
        self.y = asarray(y, dtype=float)
        # Closes the polygon if were open
        x1, y1 = x[0], y[0]
        xn, yn = x[-1], y[-1]
        if x1 != xn or y1 != yn:
            self.x = concatenate((self.x, [x1]))
            self.y = concatenate((self.y, [y1]))
        # Anti-clockwise coordinates
        if _det(self.x, self.y) < 0:
            self.x = self.x[::-1]
            self.y = self.y[::-1]

    def is_inside(self, xpoint, ypoint, smalld=1e-12):
        """Check if point is inside a general polygon.

        Input parameters:
        xpoint -- The x-coord of the point to be tested.
        ypoint -- The y-coords of the point to be tested.
        smalld -- A small float number.

        xpoint and ypoint could be scalars or array-like sequences.

        Output parameters:

        mindst -- The distance from the point to the nearest point of the
                  polygon.
                  If mindst < 0 then point is outside the polygon.
                  If mindst = 0 then point in on a side of the polygon.
                  If mindst > 0 then point is inside the polygon.

        Notes
        -----
        An improved version of the algorithm of Nordbeck and Rydstedt.
        REF: SLOAN, S.W. (1985): A point-in-polygon program. Adv. Eng.
             Software, Vol 7, No. 1, pp 45-47.

        """
        xpoint = asarray(xpoint, dtype=float)
        ypoint = asarray(ypoint, dtype=float)
        # Scalar to array
        if xpoint.shape == ():
            xpoint = array([xpoint], dtype=float)
            ypoint = array([ypoint], dtype=float)
            scalar = True
        else:
            scalar = False
        # Check consistency
        if xpoint.shape != ypoint.shape:
            msg = 'x and y has different shapes'
            raise IndexError(msg)
        # If snear = True: Dist to nearest side < nearest vertex
        # If snear = False: Dist to nearest vertex < nearest side
        snear = ma.masked_all(xpoint.shape, dtype=bool)
        # Initialize arrays
        mindst = ones_like(xpoint, dtype=float) * inf
        j = ma.masked_all(xpoint.shape, dtype=int)
        x = self.x
        y = self.y
        n = len(x) - 1  # Number of sides/vertices defining the polygon
        # Loop over each side defining polygon
        for i in range(n):
            d = ones_like(xpoint, dtype=float) * inf
            # Start of side has coords (x1, y1)
            # End of side has coords (x2, y2)
            # Point has coords (xpoint, ypoint)
            x1 = x[i]
            y1 = y[i]
            x21 = x[i + 1] - x1
            y21 = y[i + 1] - y1
            x1p = x1 - xpoint
            y1p = y1 - ypoint
            # Points on infinite line defined by
            #     x = x1 + t * (x1 - x2)
            #     y = y1 + t * (y1 - y2)
            # where
            #     t = 0    at (x1, y1)
            #     t = 1    at (x2, y2)
            # Find where normal passing through (xpoint, ypoint) intersects
            # infinite line
            t = -(x1p * x21 + y1p * y21) / (x21**2 + y21**2)
            tlt0 = t < 0
            tle1 = (t >= 0) & (t <= 1)
            # Normal intersects side
            d[tle1] = (x1p[tle1] + t[tle1] * x21) ** 2 + (y1p[tle1] + t[tle1] * y21) ** 2
            # Normal does not intersects side
            # Point is closest to vertex (x1, y1)
            # Compute square of distance to this vertex
            d[tlt0] = x1p[tlt0] ** 2 + y1p[tlt0] ** 2
            # Store distances
            mask = d < mindst
            mindst[mask] = d[mask]
            j[mask] = i
            # Point is closer to (x1, y1) than any other vertex or side
            snear[mask & tlt0] = False
            # Point is closer to this side than to any other side or vertex
            snear[mask & tle1] = True
        if ma.count(snear) != snear.size:
            msg = 'Error computing distances'
            raise IndexError(msg)
        mindst **= 0.5
        # Point is closer to its nearest vertex than its nearest side, check if
        # nearest vertex is concave.
        # If the nearest vertex is concave then point is inside the polygon,
        # else the point is outside the polygon.
        jo = j.copy()
        jo[j == 0] -= 1
        area = _det([x[j + 1], x[j], x[jo - 1]], [y[j + 1], y[j], y[jo - 1]])
        mindst[~snear] = copysign(mindst, area)[~snear]
        # Point is closer to its nearest side than to its nearest vertex, check
        # if point is to left or right of this side.
        # If point is to left of side it is inside polygon, else point is
        # outside polygon.
        area = _det([x[j], x[j + 1], xpoint], [y[j], y[j + 1], ypoint])
        mindst[snear] = copysign(mindst, area)[snear]
        # Point is on side of polygon
        mindst[fabs(mindst) < smalld] = 0
        # If input values were scalar then the output should be too
        if scalar:
            mindst = float(mindst)
        return mindst


@deprecated_alias({'gpos': 'pos'})
class Grid(ABCHasStrictTraits):
    """Virtual base class for grid geometries.

    Defines the common interface for all grid classes and
    provides facilities to query grid properties and related data. This class
    may be used as a base for specialized grid implementations. It should not
    be used directly as it contains no real functionality.
    """

    #: Overall number of grid points. Readonly; is set automatically when
    #: other grid defining properties are set
    size = Property(desc='overall number of grid points')

    #: Shape of grid. Readonly, gives the shape as tuple, useful for cartesian
    #: grids
    shape = Property(desc='grid shape as tuple')

    #: Grid positions as (3, :attr:`size`) array of floats, without invalid
    #: microphones; readonly.
    pos = Property(desc='x, y, z positions of grid points')

    # internal identifier
    digest = Property

    @abstractmethod
    def _get_digest(self):
        """Returns the digest of the grid object."""

    # 'digest' is a placeholder for other properties in derived classes,
    # necessary to trigger the depends on mechanism
    @property_depends_on(['digest'])
    @abstractmethod
    def _get_size(self):
        """Returns the number of grid points."""

    # 'digest' is a placeholder for other properties in derived classes
    @property_depends_on(['digest'])
    @abstractmethod
    def _get_shape(self):
        """Returns the shape of the grid as a Tuple."""

    @property_depends_on(['digest'])
    @abstractmethod
    def _get_pos(self):
        """Returns the grid positions as (3, size) array of floats."""

    def subdomain(self, sector):
        """Queries the indices for a subdomain in the grid.

        Allows arbitrary subdomains of type :class:`Sector`

        Parameters
        ----------
        sector : :class:`Sector`
            Sector describing the subdomain.

        Returns
        -------
        2-tuple of arrays of integers or of numpy slice objects
            The indices that can be used to mask/select the grid subdomain from
            an array with the same shape as the grid.

        """
        xpos = self.pos
        # construct grid-shaped array with "True" entries where sector is
        xyi = sector.contains(xpos).reshape(self.shape)
        # return indices of "True" entries
        return where(xyi)


@deprecated_alias({'gpos': 'pos'}, read_only=True)
class RectGrid(Grid):
    """Provides a cartesian 2D grid for the beamforming results.

    The grid has square or nearly square cells and is on a plane perpendicular
    to the z-axis. It is defined by lower and upper x- and  y-limits and the
    z co-ordinate.
    """

    #: The lower x-limit that defines the grid, defaults to -1.
    x_min = Float(-1.0, desc='minimum  x-value')

    #: The upper x-limit that defines the grid, defaults to 1.
    x_max = Float(1.0, desc='maximum  x-value')

    #: The lower y-limit that defines the grid, defaults to -1.
    y_min = Float(-1.0, desc='minimum  y-value')

    #: The upper y-limit that defines the grid, defaults to 1.
    y_max = Float(1.0, desc='maximum  y-value')

    #: The z co-ordinate that defines the grid, defaults to 1.
    z = Float(1.0, desc='position on z-axis')

    #: The cell side length for the grid, defaults to 0.1.
    increment = Float(0.1, desc='step size')

    #: Number of grid points along x-axis, readonly.
    nxsteps = Property(desc='number of grid points along x-axis')

    #: Number of grid points along y-axis, readonly.
    nysteps = Property(desc='number of grid points along y-axis')

    # internal identifier
    digest = Property(
        depends_on=['x_min', 'x_max', 'y_min', 'y_max', 'z', 'increment'],
    )

    @property_depends_on(['nxsteps', 'nysteps'])
    def _get_size(self):
        return self.nxsteps * self.nysteps

    @property_depends_on(['nxsteps', 'nysteps'])
    def _get_shape(self):
        return (self.nxsteps, self.nysteps)

    @property_depends_on(['x_min', 'x_max', 'increment'])
    def _get_nxsteps(self):
        i = abs(self.increment)
        if i != 0:
            return int(round((abs(self.x_max - self.x_min) + i) / i))
        return 1

    @property_depends_on(['y_min', 'y_max', 'increment'])
    def _get_nysteps(self):
        i = abs(self.increment)
        if i != 0:
            return int(round((abs(self.y_max - self.y_min) + i) / i))
        return 1

    @cached_property
    def _get_digest(self):
        return digest(self)

    @property_depends_on(['x_min', 'x_max', 'y_min', 'y_max', 'increment'])
    def _get_pos(self):
        """Calculates grid co-ordinates.

        Returns
        -------
        array of floats of shape (3, :attr:`~Grid.size`)
            The grid point x, y, z-coordinates in one array.

        """
        bpos = mgrid[
            self.x_min : self.x_max : self.nxsteps * 1j,
            self.y_min : self.y_max : self.nysteps * 1j,
            self.z : self.z + 0.1,
        ]
        bpos.resize((3, self.size))
        return bpos

    def index(self, x, y):
        """Queries the indices for a grid point near a certain co-ordinate.

        This can be used to query results or co-ordinates at/near a certain
        co-ordinate.

        Parameters
        ----------
        x, y : float
            The co-ordinates for which the indices are queried.

        Returns
        -------
        2-tuple of integers
            The indices that give the grid point nearest to the given x, y
            co-ordinates from an array with the same shape as the grid.

        """
        if x < self.x_min or x > self.x_max:
            msg = 'x-value out of range'
            raise ValueError(msg)
        if y < self.y_min or y > self.y_max:
            msg = 'y-value out of range'
            raise ValueError(msg)
        xi = int((x - self.x_min) / self.increment + 0.5)
        yi = int((y - self.y_min) / self.increment + 0.5)
        return xi, yi

    def indices(self, *r):
        """Queries the indices for a subdomain in the grid.

        Allows either rectangular, circular or polygonial subdomains.
        This can be used to mask or to query results from a certain
        sector or subdomain.

        Parameters
        ----------
        x1, y1, x2, y2, ... : float
            If three parameters are given, then a circular sector is assumed
            that is given by its center (x1, y1) and the radius x2.
            If four parameters are given, then a rectangular sector is
            assumed that is given by two corners (x1, y1) and (x2, y2).
            If more parameters are given, the subdomain is assumed to have
            polygonial shape with corners at (x_n, y_n).

        Returns
        -------
        2-tuple of arrays of integers or of numpy slice objects
            The indices that can be used to mask/select the grid subdomain from
            an array with the same shape as the grid.

        """
        if len(r) == 3:  # only 3 values given -> use x,y,radius method
            xpos = self.pos
            xis = []
            yis = []
            dr2 = (xpos[0, :] - r[0]) ** 2 + (xpos[1, :] - r[1]) ** 2
            # array with true/false entries
            inds = dr2 <= r[2] ** 2
            for np in arange(self.size)[inds]:  # np -- points in x2-circle
                xi, yi = self.index(xpos[0, np], xpos[1, np])
                xis += [xi]
                yis += [yi]
            if not (xis and yis):  # if no points in circle, take nearest one
                return self.index(r[0], r[1])
            return array(xis), array(yis)
        if len(r) == 4:  # rectangular subdomain - old functionality
            xi1, yi1 = self.index(min(r[0], r[2]), min(r[1], r[3]))
            xi2, yi2 = self.index(max(r[0], r[2]), max(r[1], r[3]))
            return s_[xi1 : xi2 + 1], s_[yi1 : yi2 + 1]
        xpos = self.pos
        xis = []
        yis = []
        # replaced matplotlib Path by numpy
        # p = Path(array(r).reshape(-1,2))
        # inds = p.contains_points()
        # inds = in_poly(xpos[:2,:].T,array(r).reshape(-1,2))
        poly = Polygon(array(r).reshape(-1, 2)[:, 0], array(r).reshape(-1, 2)[:, 1])
        dists = poly.is_inside(xpos[0, :], xpos[1, :])
        inds = dists >= 0
        for np in arange(self.size)[inds]:  # np -- points in x2-circle
            xi, yi = self.index(xpos[0, np], xpos[1, np])
            xis += [xi]
            yis += [yi]
        if not (xis and yis):  # if no points inside, take nearest to center
            center = array(r).reshape(-1, 2).mean(0)
            return self.index(center[0], center[1])
        return array(xis), array(yis)
        # return arange(self.size)[inds]

    def extend(self):
        """The extension of the grid in pylab.imshow compatible form.

        Returns
        -------
        4-tuple of floats
            The extent of the grid as a tuple of x_min, x_max, y_min, y_max)

        """
        return (self.x_min, self.x_max, self.y_min, self.y_max)


class RectGrid3D(RectGrid):
    """Provides a cartesian 3D grid for the beamforming results.

    The grid has cubic or nearly cubic cells. It is defined by lower and upper
    x-, y- and  z-limits.
    """

    #: The lower z-limit that defines the grid, defaults to -1.
    z_min = Float(-1.0, desc='minimum  z-value')

    #: The upper z-limit that defines the grid, defaults to 1.
    z_max = Float(1.0, desc='maximum  z-value')

    #: Number of grid points along x-axis, readonly.
    nxsteps = Property(desc='number of grid points along x-axis')

    #: Number of grid points along y-axis, readonly.
    nysteps = Property(desc='number of grid points along y-axis')

    #: Number of grid points along x-axis, readonly.
    nzsteps = Property(desc='number of grid points along x-axis')

    # Private trait for increment handling
    _increment = Union(Float(), CArray(shape=(3,), dtype=float), default_value=0.1, desc='step size')

    #: The cell side length for the grid. This can either be a scalar (same
    #: increments in all 3 dimensions) or a (3,) array of floats with
    #: respective increments in x,y, and z-direction (in m).
    #: Defaults to 0.1.
    increment = Property(desc='step size')

    def _get_increment(self):
        return self._increment

    def _set_increment(self, increment):
        if isscalar(increment):
            try:
                self._increment = absolute(float(increment))
            except ValueError as ve:
                raise TraitError(args=self, name='increment', info='Float or CArray(3,)', value=increment) from ve
        elif len(increment) == 3:
            self._increment = array(increment, dtype=float)
        else:
            raise (TraitError(args=self, name='increment', info='Float or CArray(3,)', value=increment))

    # internal identifier
    digest = Property(
        depends_on=['x_min', 'x_max', 'y_min', 'y_max', 'z_min', 'z_max', '_increment'],
    )

    @property_depends_on(['nxsteps', 'nysteps', 'nzsteps'])
    def _get_size(self):
        return self.nxsteps * self.nysteps * self.nzsteps

    @property_depends_on(['nxsteps', 'nysteps', 'nzsteps'])
    def _get_shape(self):
        return (self.nxsteps, self.nysteps, self.nzsteps)

    @property_depends_on(['x_min', 'x_max', '_increment'])
    def _get_nxsteps(self):
        i = abs(self.increment) if isscalar(self.increment) else abs(self.increment[0])
        if i != 0:
            return int(round((abs(self.x_max - self.x_min) + i) / i))
        return 1

    @property_depends_on(['y_min', 'y_max', '_increment'])
    def _get_nysteps(self):
        i = abs(self.increment) if isscalar(self.increment) else abs(self.increment[1])
        if i != 0:
            return int(round((abs(self.y_max - self.y_min) + i) / i))
        return 1

    @property_depends_on(['z_min', 'z_max', '_increment'])
    def _get_nzsteps(self):
        i = abs(self.increment) if isscalar(self.increment) else abs(self.increment[2])
        if i != 0:
            return int(round((abs(self.z_max - self.z_min) + i) / i))
        return 1

    @property_depends_on('digest')
    def _get_pos(self):
        """Calculates grid co-ordinates.

        Returns
        -------
        array of floats of shape (3, :attr:`~Grid.size`)
            The grid point x, y, z-coordinates in one array.

        """
        bpos = mgrid[
            self.x_min : self.x_max : self.nxsteps * 1j,
            self.y_min : self.y_max : self.nysteps * 1j,
            self.z_min : self.z_max : self.nzsteps * 1j,
        ]
        bpos.resize((3, self.size))
        return bpos

    @cached_property
    def _get_digest(self):
        return digest(self)

    def index(self, x, y, z):
        """Queries the indices for a grid point near a certain co-ordinate.

        This can be used to query results or co-ordinates at/near a certain
        co-ordinate.

        Parameters
        ----------
        x, y, z : float
            The co-ordinates for which the indices is queried.

        Returns
        -------
        3-tuple of integers
            The indices that give the grid point nearest to the given x, y, z
            co-ordinates from an array with the same shape as the grid.

        """
        if x < self.x_min or x > self.x_max:
            msg = f'x-value out of range {x:f} ({self.x_min:f}, {self.x_max:f})'
            raise ValueError(msg)
        if y < self.y_min or y > self.y_max:
            msg = f'y-value out of range {y:f} ({self.y_min:f}, {self.y_max:f})'
            raise ValueError(msg)
        if z < self.z_min or z > self.z_max:
            msg = f'z-value out of range {z:f} ({self.z_min:f}, {self.z_max:f})'
            raise ValueError(msg)
        if isscalar(self.increment):
            incx = incy = incz = self.increment
        else:
            incx, incy, incz = self.increment
        xi = int(round((x - self.x_min) / incx))
        yi = int(round((y - self.y_min) / incy))
        zi = int(round((z - self.z_min) / incz))
        return xi, yi, zi

    def indices(self, x1, y1, z1, x2, y2, z2):
        """Queries the indices for a subdomain in the grid.

        Allows box-shaped subdomains. This can be used to
        mask or to query results from a certain sector or subdomain.

        Parameters
        ----------
        x1, y1, z1, x2, y2, z2 : float
            A box-shaped sector is assumed that is given by two corners
            (x1,y1,z1) and (x2,y2,z2).

        Returns
        -------
        3-tuple of numpy slice objects
            The indices that can be used to mask/select the grid subdomain from
            an array with the same shape as the grid.

        """
        xi1, yi1, zi1 = self.index(min(x1, x2), min(y1, y2), min(z1, z2))
        xi2, yi2, zi2 = self.index(max(x1, x2), max(y1, y2), max(z1, z2))
        return s_[xi1 : xi2 + 1], s_[yi1 : yi2 + 1], s_[zi1 : zi2 + 1]


@deprecated_alias({'from_file': 'file', 'gpos_file': 'pos'})
class ImportGrid(Grid):
    """Loads a 3D grid from xml file."""

    #: Name of the .xml-file from which to read the data.
    file = File(filter=['*.xml'], exists=True, desc='name of the xml file to import')

    _gpos = CArray(dtype=float, desc='x, y, z position of all Grid Points')

    subgrids = CArray(desc='names of subgrids for each point')

    # internal identifier
    digest = Property(depends_on=['_gpos'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    # 'digest' is a placeholder for other properties in derived classes,
    # necessary to trigger the depends on mechanism
    @property_depends_on(['_gpos'])
    def _get_size(self):
        return self.pos.shape[-1]

    # 'digest' is a placeholder for other properties in derived classes
    @property_depends_on(['_gpos'])
    def _get_shape(self):
        return (self.pos.shape[-1],)

    @property_depends_on(['_gpos'])
    def _get_pos(self):
        return self._gpos

    def _set_pos(self, pos):
        self._gpos = pos

    @on_trait_change('file')
    def import_gpos(self):
        """Import the the grid point locations from .xml file when :attr:`file` changes."""
        doc = xml.dom.minidom.parse(self.file)
        names = []
        xyz = []
        for el in doc.getElementsByTagName('pos'):
            names.append(el.getAttribute('subgrid'))
            xyz.append([float(el.getAttribute(a)) for a in 'xyz'])
        self._gpos = array(xyz, 'd').swapaxes(0, 1)
        self.subgrids = array(names)


@deprecated_alias({'gpos': 'pos', 'numpoints': 'num_points'}, read_only=['gpos'])
class LineGrid(Grid):
    """Class for Line grid geometries."""

    #: Staring point of the Grid
    loc = Tuple((0.0, 0.0, 0.0))

    #: Vector to define the orientation of the line source
    direction = Tuple((1.0, 0.0, 0.0), desc='Line orientation ')

    #: Vector to define the length of the line source in meter
    length = Float(1, desc='length of the line source')

    #:number of grid points.
    num_points = Int(1, desc='length of the line source')

    #: Overall number of grid points. Readonly; is set automatically when
    #: other grid defining properties are set
    size = Property(desc='overall number of grid points')

    #: Grid positions as (3, :attr:`size`) array of floats, without invalid
    #: microphones; readonly.
    pos = Property(desc='x, y, z positions of grid points')

    digest = Property(
        depends_on=['loc', 'direction', 'length', 'num_points', 'size'],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    # 'digest' is a placeholder for other properties in derived classes,
    # necessary to trigger the depends on mechanism
    @property_depends_on(['num_points'])
    def _get_size(self):
        return self.pos.shape[-1]

    # 'digest' is a placeholder for other properties in derived classes
    @property_depends_on(['num_points'])
    def _get_shape(self):
        return self.pos.shape[-1]

    @property_depends_on(['num_points', 'length', 'direction', 'loc'])
    def _get_pos(self):
        dist = self.length / (self.num_points - 1)
        loc = array(self.loc, dtype=float).reshape((3, 1))
        direc_n = array(self.direction) / norm(self.direction)
        pos = zeros((self.num_points, 3))
        for s in range(self.num_points):
            pos[s] = loc.T + direc_n * dist * s
        return pos.T


@deprecated_alias({'gpos': 'pos'}, read_only=True)
class MergeGrid(Grid):
    """Base class for merging different grid geometries."""

    #: List of Grids to be merged
    #: each grid gets a new subdomain in the new grid
    #: other grid defining properties are set
    grids = List(desc='list of grids')

    grid_digest = Str(desc='digest of the merged grids')

    subgrids = Property(desc='names of subgrids for each point')

    # internal identifier
    digest = Property(
        depends_on=['grids', 'grid_digest'],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    @observe('grids.items.digest')
    def _set_sourcesdigest(self, event):  # noqa ARG002
        self.grid_digest = ldigest(self.grids)

    # 'digest' is a placeholder for other properties in derived classes,
    # necessary to trigger the depends on mechanism
    @property_depends_on(['digest'])
    def _get_size(self):
        return self.pos.shape[-1]

    # 'digest' is a placeholder for other properties in derived classes
    @property_depends_on(['digest'])
    def _get_shape(self):
        return self.pos.shape[-1]

    @property_depends_on(['digest'])
    def _get_subgrids(self):
        subgrids = zeros((1, 0), dtype=str)
        for grid in self.grids:
            subgrids = append(subgrids, tile(grid.__class__.__name__ + grid.digest, grid.size))
        return subgrids[:, newaxis].T

    @property_depends_on(['digest'])
    def _get_pos(self):
        bpos = zeros((3, 0))
        for grid in self.grids:
            bpos = append(bpos, grid.pos, axis=1)
        return unique(bpos, axis=1)


class Sector(ABCHasStrictTraits):
    """Base class for all sector types.

    Defines the common interface for all tbdsector classes. This class
    may be used as a base for diverse sector implementations. If used
    directly, it implements a sector encompassing the whole grid.
    """

    def contains(self, pos):
        """Queries whether the coordinates in a given array lie within the
        defined sector.
        For this sector type, any position is valid.

        Parameters
        ----------
        pos : array of floats
            Array with the shape 3x[number of gridpoints] containing the
            grid positions

        Returns
        -------
        array of bools with as many entries as columns in pos
            Array indicating which of the given positions lie within the
            given sector

        """
        return ones(pos.shape[1], dtype=bool)


class SingleSector(Sector):
    """Base class for single sector types.

    Defines the common interface for all single sector classes. This class
    may be used as a base for diverse single sector implementations. If used
    directly, it implements a sector encompassing the whole grid.
    """

    #: Boolean flag, if 'True' (default), grid points lying on the sector border are included.
    include_border = Bool(True, desc='include points on the border')

    #: Absolute tolerance for sector border
    abs_tol = Float(1e-12, desc='absolute tolerance for sector border')

    #: Boolean flag, if 'True' (default), the nearest grid point is returned if None is inside the
    #: sector.
    default_nearest = Bool(True, desc='return nearest grid point to center of none inside sector')


class RectSector(SingleSector):
    """Class for defining a rectangular sector.

    Can be used for 2D Grids for defining a rectangular sector or
    for 3D grids for a rectangular cylinder sector parallel to the z-axis.
    """

    #: The lower x position of the rectangle
    x_min = Float(-1.0, desc='minimum x position of the rectangle')

    #: The upper x position of the rectangle
    x_max = Float(1.0, desc='maximum x position of the rectangle')

    #: The lower y position of the rectangle
    y_min = Float(-1.0, desc='minimum y position of the rectangle')

    #: The upper y position of the rectangle
    y_max = Float(1.0, desc='maximum y position of the rectangle')

    def contains(self, pos):
        """Queries whether the coordinates in a given array lie within the
        rectangular sector.
        If no coordinate is inside, the nearest one to the rectangle center
        is returned if :attr:`~Sector.default_nearest` is True.

        Parameters
        ----------
        pos : array of floats
            Array with the shape 3x[number of gridpoints] containing the
            grid positions

        Returns
        -------
        array of bools with as many entries as columns in pos
            Array indicating which of the given positions lie within the
            given sector

        """
        # make sure xmin is minimum etc
        xmin = min(self.x_min, self.x_max)
        xmax = max(self.x_min, self.x_max)
        ymin = min(self.y_min, self.y_max)
        ymax = max(self.y_min, self.y_max)

        abs_tol = self.abs_tol
        # get pos indices inside rectangle (* == and)
        if self.include_border:
            inds = (
                (pos[0, :] - xmin > -abs_tol)
                * (pos[0, :] - xmax < abs_tol)
                * (pos[1, :] - ymin > -abs_tol)
                * (pos[1, :] - ymax < abs_tol)
            )
        else:
            inds = (
                (pos[0, :] - xmin > abs_tol)
                * (pos[0, :] - xmax < -abs_tol)
                * (pos[1, :] - ymin > abs_tol)
                * (pos[1, :] - ymax < -abs_tol)
            )

        # if none inside, take nearest
        if ~inds.any() and self.default_nearest:
            x = (xmin + xmax) / 2.0
            y = (ymin + ymax) / 2.0
            dr2 = (pos[0, :] - x) ** 2 + (pos[1, :] - y) ** 2
            inds[argmin(dr2)] = True

        return inds.astype(bool)


class RectSector3D(RectSector):
    """Class for defining a cuboid sector.

    Can be used for 3D Grids for defining a cuboid sector.
    """

    #: The lower z position of the cuboid
    z_min = Float(-1.0, desc='minimum z position of the cuboid')

    #: The upper z position of the cuboid
    z_max = Float(1.0, desc='maximum z position of the cuboid')

    def contains(self, pos):
        """Queries whether the coordinates in a given array lie within the
        rectangular sector.
        If no coordinate is inside, the nearest one to the rectangle center
        is returned if :attr:`~Sector.default_nearest` is True.

        Parameters
        ----------
        pos : array of floats
            Array with the shape 3x[number of gridpoints] containing the
            grid positions

        Returns
        -------
        array of bools with as many entries as columns in pos
            Array indicating which of the given positions lie within the
            given sector

        """
        # make sure xmin is minimum etc
        xmin = min(self.x_min, self.x_max)
        xmax = max(self.x_min, self.x_max)
        ymin = min(self.y_min, self.y_max)
        ymax = max(self.y_min, self.y_max)
        zmin = min(self.z_min, self.z_max)
        zmax = max(self.z_min, self.z_max)

        abs_tol = self.abs_tol
        # get pos indices inside rectangle (* == and)
        if self.include_border:
            inds = (
                (pos[0, :] - xmin > -abs_tol)
                * (pos[0, :] - xmax < abs_tol)
                * (pos[1, :] - ymin > -abs_tol)
                * (pos[1, :] - ymax < abs_tol)
                * (pos[2, :] - zmin > -abs_tol)
                * (pos[2, :] - zmax < abs_tol)
            )
        else:
            inds = (
                (pos[0, :] - xmin > abs_tol)
                * (pos[0, :] - xmax < -abs_tol)
                * (pos[1, :] - ymin > abs_tol)
                * (pos[1, :] - ymax < -abs_tol)
                * (pos[2, :] - zmin > abs_tol)
                * (pos[2, :] - zmax < -abs_tol)
            )

        # if none inside, take nearest
        if ~inds.any() and self.default_nearest:
            x = (xmin + xmax) / 2.0
            y = (ymin + ymax) / 2.0
            dr2 = (pos[0, :] - x) ** 2 + (pos[1, :] - y) ** 2
            inds[argmin(dr2)] = True

        return inds.astype(bool)


class CircSector(SingleSector):
    """Class for defining a circular sector.

    Can be used for 2D Grids for defining a circular sector or
    for 3D grids for a cylindrical sector parallel to the z-axis.
    """

    #: x position of the circle center
    x = Float(0.0, desc='x position of the circle center')

    #: y position of the circle center
    y = Float(0.0, desc='y position of the circle center')

    #: radius of the circle
    r = Float(1.0, desc='radius of the circle')

    def contains(self, pos):
        """Queries whether the coordinates in a given array lie within the
        circular sector.
        If no coordinate is inside, the nearest one outside is returned
        if :attr:`~Sector.default_nearest` is True.

        Parameters
        ----------
        pos : array of floats
            Array with the shape 3x[number of gridpoints] containing the
            grid positions

        Returns
        -------
        array of bools with as many entries as columns in pos
            Array indicating which of the given positions lie within the
            given sector

        """
        dr2 = (pos[0, :] - self.x) ** 2 + (pos[1, :] - self.y) ** 2
        # which points are in the circle?
        inds = dr2 - self.r**2 < self.abs_tol if self.include_border else dr2 - self.r**2 < -self.abs_tol

        # if there's no point inside
        if ~inds.any() and self.default_nearest:
            inds[argmin(dr2)] = True

        return inds


class PolySector(SingleSector):
    """Class for defining a polygon sector.

    Can be used for 2D Grids for defining a polygon sector.
    """

    # x1, y1, x2, y2, ... xn, yn :
    edges = List(Float)

    def contains(self, pos):
        """Queries whether the coordinates in a given array lie within the polygon sector. If no
        coordinate is inside, the nearest one to the rectangle center is returned if
        :attr:`~Sector.default_nearest` is True.

        Parameters
        ----------
        pos : array of floats
            Array with the shape 3x[number of gridpoints] containing the
            grid positions

        Returns
        -------
        array of bools with as many entries as columns in pos
            Array indicating which of the given positions lie within the
            given sector

        """
        poly = Polygon(array(self.edges).reshape(-1, 2)[:, 0], array(self.edges).reshape(-1, 2)[:, 1])
        dists = poly.is_inside(pos[0, :], pos[1, :])
        inds = dists >= -self.abs_tol if self.include_border else dists > 0

        # if none inside, take nearest
        if ~inds.any() and self.default_nearest:
            dr2 = array(self.edges).reshape(-1, 2).mean(0)
            inds[argmin(dr2)] = True

        return inds


class ConvexSector(SingleSector):
    """Class for defining a convex hull sector.

    Can be used for 2D Grids for defining a convex hull sector.
    """

    # x1, y1, x2, y2, ... xn, yn :
    edges = List(Float)

    def contains(self, pos):
        """Queries whether the coordinates in a given array lie within the
        convex sector.
        If no coordinate is inside, the nearest one to the rectangle center
        is returned if :attr:`~Sector.default_nearest` is True.

        Parameters
        ----------
        pos : array of floats
            Array with the shape 3x[number of gridpoints] containing the
            grid positions

        Returns
        -------
        array of bools with as many entries as columns in pos
            Array indicating which of the given positions lie within the
            given sector

        """
        inds = in_hull(pos[:2, :].T, array(self.edges).reshape(-1, 2), border=self.include_border, tol=self.abs_tol)

        # if none inside, take nearest
        if ~inds.any() and self.default_nearest:
            dr2 = array(self.edges).reshape(-1, 2).mean(0)
            inds[argmin(dr2)] = True

        return inds


class MultiSector(Sector):
    """Class for defining a sector consisting of multiple sectors.

    Can be used to sum over different sectors. Takes a list of sectors
    and returns the points contained in each sector.

    """

    #: List of :class:`acoular.grids.Sector` objects
    #: to be mixed.
    sectors = List(Instance(Sector))

    def contains(self, pos):
        """Queries whether the coordinates in a given array lie within any
        of the sub-sectors.

        Parameters
        ----------
        pos : array of floats
            Array with the shape 3x[number of gridpoints] containing the
            grid positions

        Returns
        -------
        array of bools with as many entries as columns in pos
            Array indicating which of the given positions lie within the
            sectors

        """
        # initialize with only "False" entries
        inds = zeros(pos.shape[1], dtype=bool)

        # add points contained in each sector
        for sec in self.sectors:
            inds += sec.contains(pos)

        return inds.astype(bool)



# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------
"""Implements acoustic environments with and without flow.

.. autosummary::
    :toctree: generated/

    Environment
    UniformFlowEnvironment
    GeneralFlowEnvironment
    FlowField
    OpenJet
    RotatingFlow
    SlotJet

"""

from abc import abstractmethod

import numba as nb
from numpy import (
    arange,
    arccos,
    arctan2,
    array,
    ascontiguousarray,
    cos,
    cross,
    dot,
    empty,
    exp,
    float32,
    float64,
    hstack,
    identity,
    isscalar,
    matmul,
    newaxis,
    pi,
    sign,
    sin,
    sqrt,
    sum,  # noqa: A004
    vstack,
    zeros_like,
)
from scipy.integrate import ode
from scipy.interpolate import LinearNDInterpolator
from scipy.linalg import norm
from scipy.spatial import ConvexHull
from traits.api import (
    ABCHasStrictTraits,
    CArray,
    Dict,
    Float,
    HasStrictTraits,
    Instance,
    Int,
    Property,
    Union,
    cached_property,
)

from .internal import digest

f64ro = nb.types.Array(nb.types.float64, 2, 'C', readonly=True)
f32ro = nb.types.Array(nb.types.float32, 2, 'C', readonly=True)


@nb.njit([(f64ro, f64ro), (f64ro, f32ro), (f32ro, f64ro), (f32ro, f32ro)], cache=True, fastmath=True)
def dist_mat(gpos, mpos):  # pragma: no cover
    """Computes distance matrix, accelerated with numba.

    Args:
    ----
        gpos (3,N)
        mpos (3,M)

    Returns
    -------
        (N,M) distance matrix

    """
    _, M = mpos.shape
    _, N = gpos.shape
    rm = empty((N, M), dtype=gpos.dtype)
    TWO = rm.dtype.type(2.0)  # make sure to have a float32 or float 64 literal
    m0 = mpos[0]
    m1 = mpos[1]
    m2 = mpos[2]
    for n in range(N):
        g0 = gpos[0, n]
        g1 = gpos[1, n]
        g2 = gpos[2, n]
        for m in range(M):
            rm[n, m] = sqrt((g0 - m0[m]) ** TWO + (g1 - m1[m]) ** TWO + (g2 - m2[m]) ** TWO)
    return rm


def cartToCyl(x, Q=None):  # noqa: N802, N803
    """Returns the cylindrical coordinate representation of a input position
    which was before transformed into a modified cartesian coordinate, which
    has flow into positive z direction.

    Parameters
    ----------
    x : float[3, nPoints]
        cartesian coordinates of n points
    Q : float[3,3]
        Orthogonal transformation matrix. If provided, the pos vectors are
        transformed via posiMod = Q * x, before transforming those modified
        coordinates into cylindrical ones. Default is identity matrix.

    Returns
    -------
    cylCoord : [3, nPoints]
        cylindrical representation of those n points with (phi, r, z)

    """
    Q = identity(3) if Q is None else Q
    if not (Q == identity(3)).all():  # noqa: SIM300
        x = matmul(Q, x)  # modified position vector
    return array([arctan2(x[1], x[0]), sqrt(x[0] ** 2 + x[1] ** 2), x[2]])


def cylToCart(x, Q=None):  # noqa: N802, N803
    """Returns the cartesian coordinate representation of a input position
    which was before transformed into a cylindrical coordinate, which
    has flow into positive z direction.

    Parameters
    ----------
    x : float[3, nPoints]
        cylindrical representation of those n points with (phi, r, z)
        cartesian coordinates of n points

    Q : float[3,3]
    Orthogonal transformation matrix. If provided, the pos vectors are
    transformed via posiMod = Q * x, before transforming those modified
    coordinates into cylindrical ones. Default is identity matrix.


    Returns
    -------
    CartCoord : [3, nPoints]
    cartesian coordinates of n points

    """
    Q = identity(3) if Q is None else Q
    if not (Q == identity(3)).all():  # noqa: SIM300
        x = matmul(Q, x)  # modified position vector
    return array([x[1] * sin(x[0]), x[1] * cos(x[0]), x[2]])


class Environment(HasStrictTraits):
    """A simple acoustic environment without flow.

    This class provides the facilities to calculate the travel time (distances)
    between grid point locations and microphone locations.
    """

    # internal identifier
    digest = Property(
        depends_on=['c'],
    )

    #: The speed of sound, defaults to 343 m/s
    c = Float(343.0, desc='speed of sound')

    #: The region of interest (ROI), not needed for most types of environment
    roi = Union(None, CArray)

    def _get_digest(self):
        return digest(self)

    def _r(self, gpos, mpos=0.0):
        """Calculates distances between grid point locations and microphone
        locations or the origin. Functionality may change in the future.

        Parameters
        ----------
        gpos : array of floats of shape (3, N)
            The locations of points in the beamforming map grid in 3D cartesian
            co-ordinates.
        mpos : array of floats of shape (3, M), optional
            The locations of microphones in 3D cartesian co-ordinates. If not
            given, then only one microphone at the origin (0, 0, 0) is
            considered.

        Returns
        -------
        r : array of floats
            The distances in a twodimensional (N, M) array of floats. If M==1,
            then only a one-dimensional array is returned.

        """
        if isscalar(mpos):
            mpos = array((0, 0, 0), dtype=float64)[:, newaxis]
        rm = dist_mat(ascontiguousarray(gpos), ascontiguousarray(mpos))
        #        mpos = mpos[:, newaxis, :]
        #        rmv = gpos[:, :, newaxis]-mpos
        #        rm = sum(rmv*rmv, 0)**0.5
        if rm.shape[1] == 1:
            rm = rm[:, 0]
        return rm


class UniformFlowEnvironment(Environment):
    """An acoustic environment with uniform flow.

    This class provides the facilities to calculate the travel time (distances)
    between grid point locations and microphone locations in a uniform flow
    field.
    """

    #: The Mach number, defaults to 0.
    ma = Float(0.0, desc='flow mach number')

    #: The unit vector that gives the direction of the flow, defaults to
    #: flow in x-direction.
    fdv = CArray(dtype=float64, shape=(3,), value=array((1.0, 0, 0)), desc='flow direction')

    # internal identifier
    digest = Property(
        depends_on=['c', 'ma', 'fdv'],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    def _r(self, gpos, mpos=0.0):
        """Calculates the virtual distances between grid point locations and
        microphone locations or the origin. These virtual distances correspond
        to travel times of the sound. Functionality may change in the future.

        Parameters
        ----------
        gpos : array of floats of shape (3, N)
            The locations of points in the beamforming map grid in 3D cartesian
            co-ordinates.
        mpos : array of floats of shape (3, M), optional
            The locations of microphones in 3D cartesian co-ordinates. If not
            given, then only one microphone at the origin (0, 0, 0) is
            considered.

        Returns
        -------
        array of floats
            The distances in a twodimensional (N, M) array of floats. If M==1,
            then only a one-dimensional array is returned.

        """
        if isscalar(mpos):
            mpos = array((0, 0, 0), dtype=float32)[:, newaxis]
        fdv = self.fdv / sqrt((self.fdv * self.fdv).sum())
        mpos = mpos[:, newaxis, :]
        rmv = gpos[:, :, newaxis] - mpos
        rm = sqrt(sum(rmv * rmv, 0))
        macostheta = (self.ma * sum(rmv.reshape((3, -1)) * fdv[:, newaxis], 0) / rm.reshape(-1)).reshape(rm.shape)
        rm *= 1 / (-macostheta + sqrt(macostheta * macostheta - self.ma * self.ma + 1))
        if rm.shape[1] == 1:
            rm = rm[:, 0]
        return rm


class FlowField(ABCHasStrictTraits):
    """An abstract base class for a spatial flow field."""

    digest = Property

    @abstractmethod
    def _get_digest(self):
        pass

    @abstractmethod
    def v(self, xx):  # noqa: ARG002
        """Provides the flow field as a function of the location.

        Parameters
        ----------
        xx : array of floats of shape (3, )
            Location in the fluid for which to provide the data.

        Returns
        -------
        tuple
            A tuple with two elements:
            - velocity_vector : array of floats of shape (3, )
                The velocity vector at the given location.
            - jacobian_matrix : array of floats of shape (3, 3)
                The Jacobian matrix of the velocity vector field at the given location.
        """


class SlotJet(FlowField):
    """Provides an analytical approximation of the flow field of a slot jet.

    See :cite:`Albertson1950` for details.
    """

    #: Exit velocity at jet origin, i.e. the nozzle. Defaults to 0.
    v0 = Float(0.0, desc='exit velocity')

    #: Location of a point at the slot center line,
    #: defaults to the co-ordinate origin.
    origin = CArray(dtype=float64, shape=(3,), value=array((0.0, 0.0, 0.0)), desc='center of nozzle')

    #: Unit flow direction of the slot jet, defaults to (1,0,0).
    flow = CArray(dtype=float64, shape=(3,), value=array((1.0, 0.0, 0.0)), desc='flow direction')

    #: Unit vector parallel to slot center plane, defaults to (0,1,0).
    plane = CArray(dtype=float64, shape=(3,), value=array((0.0, 1.0, 0.0)), desc='slot center line direction')

    #: Width of the slot, defaults to 0.2 .
    B = Float(0.2, desc='nozzle diameter')

    #: Nondimensional length of zone of flow establishment (jet core length), defaults to 5.2
    l = Float(5.2, desc='flow establishment length')  # noqa: E741

    # internal identifier
    digest = Property(
        depends_on=['v0', 'origin', 'flow', 'plane', 'B', 'l'],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    def v(self, xx):
        """Provides the flow field as a function of the location. This is
        implemented here only for the component in the direction of :attr:`flow`;
        entrainment components are set to zero.

        Parameters
        ----------
        xx : array of floats of shape (3, )
            Location in the fluid for which to provide the data.

        Returns
        -------
        tuple with two elements
            The first element in the tuple is the velocity vector and the
            second is the Jacobian of the velocity vector field, both at the
            given location.

        """
        # normalize
        flow = self.flow / norm(self.flow)
        plane = self.plane / norm(self.plane)
        # additional axes of global co-ordinate system
        yy = -cross(flow, plane)
        zz = cross(flow, yy)
        # distance from slot exit plane
        xx1 = xx - self.origin
        # local co-ordinate system
        x = dot(flow, xx1)
        y = dot(yy, xx1)
        x1 = 0.5668 / self.l * x  # C1 in Albertson1950
        h1 = abs(y) + sqrt(pi) * 0.5 * x1 - 0.5 * self.B
        if h1 < 0.0:
            # core jet
            Ux = self.v0
            Udx = 0
            Udy = 0
        else:
            # shear layer
            Ux = self.v0 * exp(-h1 * h1 / (2 * x1 * x1))
            Udx = (h1 * h1 / (x * x1 * x1) - sqrt(pi) * 0.5 * h1 / (x * x1)) * Ux
            Udy = -sign(y) * h1 * Ux / (x1 * x1)
        # Jacobi matrix
        dU = array(((Udx, 0, 0), (Udy, 0, 0), (0, 0, 0))).T
        # rotation matrix
        R = array((flow, yy, zz)).T
        return dot(R, array((Ux, 0, 0))), dot(dot(R, dU), R.T)


class OpenJet(FlowField):
    """Provides an analytical approximation of the flow field of an open jet.

    See :cite:`Albertson1950` for details.

    Notes
    -----
    This is not a fully generic implementation and is limited to flow in the
    x-direction only. No other directions are possible at the moment and flow
    components in the other direction are zero.

    """

    #: Exit velocity at jet origin, i.e. the nozzle. Defaults to 0.
    v0 = Float(0.0, desc='exit velocity')

    #: Location of the nozzle center, defaults to the co-ordinate origin.
    origin = CArray(dtype=float64, shape=(3,), value=array((0.0, 0.0, 0.0)), desc='center of nozzle')

    #: Diameter of the nozzle, defaults to 0.2 .
    D = Float(0.2, desc='nozzle diameter')

    #: Nondimensional length of zone of flow establishment (jet core length), defaults to 6.2
    l = Float(6.2, desc='flow establishment length')  # noqa: E741

    # internal identifier
    digest = Property(
        depends_on=['v0', 'origin', 'D', 'l'],
    )

    @cached_property
    def _get_digest(self):
        return digest(self)

    def v(self, xx):
        """Provides the flow field as a function of the location. This is
        implemented here only for a jet in `x`-direction and the `y`- and
        `z`-components are set to zero.

        Parameters
        ----------
        xx : array of floats of shape (3, )
            Location in the fluid for which to provide the data.

        Returns
        -------
        tuple with two elements
            The first element in the tuple is the velocity vector and the
            second is the Jacobian of the velocity vector field, both at the
            given location.

        """
        x, y, z = xx - self.origin
        r = sqrt(y * y + z * z)
        x1 = 0.5022 / self.l * x  # C2 in Albertson1950
        h1 = r + x1 - 0.5 * self.D
        U = self.v0 * exp(-h1 * h1 / (2 * x1 * x1))
        if h1 < 0.0:
            Udr = 0.0
            U = self.v0
        else:
            Udr = -h1 * U / (x1 * x1)
        if r > 0.0:
            Udy = y * Udr / r
            Udz = z * Udr / r
        else:
            Udy = Udz = 0.0
        Udx = (h1 * h1 / (x * x1 * x1) - h1 / (x * x1)) * U
        if h1 < 0.0:
            Udx = 0

        # flow field
        v = array((U, 0.0, 0.0))
        # Jacobi matrix
        dv = array(((Udx, 0.0, 0.0), (Udy, 0.0, 0.0), (Udz, 0.0, 0.0))).T
        return v, dv


class RotatingFlow(FlowField):
    """Analytical approximation of the flow field of a rotating fluid with constant flow."""

    #: Exit velocity at jet origin, i.e. the nozzle. Defaults to 0.
    rpm = Float(0.0, desc='revolutions per minute of the virtual array; negative values for clockwise rotation')

    v0 = Float(0.0, desc='flow velocity')

    #: Location of the nozzle center, defaults to the co-ordinate origin.
    origin = CArray(dtype=float64, shape=(3,), value=array((0.0, 0.0, 0.0)), desc='center of nozzle')

    # internal identifier
    digest = Property(
        depends_on=['v0', 'origin', 'rpm'],
    )

    # internal identifier
    omega = Property(
        depends_on=['rpm'],
    )

    @cached_property
    def _get_omega(self):
        return 2 * pi * self.rpm / 60

    @cached_property
    def _get_digest(self):
        return digest(self)

    def v(self, xx):
        """Provides the rotating flow field around the z-Axis as a function of the location.

        Parameters
        ----------
        xx : array of floats of shape (3, )
            Location in the fluid for which to provide the data.

        Returns
        -------
        tuple with two elements
            The first element in the tuple is the velocity vector and the
            second is the Jacobian of the velocity vector field, both at the
            given location.

        """
        x, y, z = xx - self.origin

        # rotational speed
        omega = self.omega

        # velocity vector
        U = omega * y
        V = -omega * x
        W = self.v0

        # flow field
        v = array((U, V, W))
        # Jacobi matrix
        dv = array(((0, -omega, 0.0), (omega, 0, 0.0), (0.0, 0.0, 0.0))).T
        return v, dv


def spiral_sphere(N, Om=None, b=None):  # noqa: N803 # change to 4*pi
    """Internal helper function for the raycasting that returns an array of
    unit vectors (N, 3) giving equally distributed directions on a part of
    sphere given by the center direction b and the solid angle Om.
    """
    Om = 2 * pi if Om is None else Om
    b = array((0, 0, 1)) if b is None else b
    # first produce 'equally' distributed directions in spherical coords
    o = 4 * pi / Om
    h = -1 + 2 * arange(N) / (N * o - 1.0)
    theta = arccos(h)
    phi = zeros_like(theta)
    for i, hk in enumerate(h[1:]):
        phi[i + 1] = phi[i] + 3.6 / sqrt(N * o * (1 - hk * hk)) % (2 * pi)
    # translate to cartesian coords
    xyz = vstack((sin(theta) * cos(phi), sin(theta) * sin(phi), cos(theta)))
    # mirror everything on a plane so that b points into the center
    a = xyz[:, 0]
    b = b / norm(b)
    ab = (a - b)[:, newaxis]
    if norm(ab) < 1e-10:
        return xyz
    # this is the Householder matrix for mirroring
    H = identity(3) - dot(ab, ab.T) / dot(ab.T, a)
    # actual mirroring
    return dot(H, xyz)


class GeneralFlowEnvironment(Environment):
    """An acoustic environment with a generic flow field.

    This class provides the facilities to calculate the travel time (distances)
    between grid point locations and microphone locations in a generic flow
    field with non-uniform velocities that depend on the location. The
    algorithm for the calculation uses a ray-tracing approach that bases on
    rays cast from every microphone position in multiple directions and traced
    backwards in time. The result is interpolated within a tetrahedal grid
    spanned between these rays.
    """

    #: The flow field, must be of type :class:`~acoular.environments.FlowField`.
    ff = Instance(FlowField, desc='flow field')

    #: Number of rays used per solid angle :math:`\Omega`, defaults to 200.
    N = Int(200, desc='number of rays per Om')

    #: The maximum solid angle used in the algorithm, defaults to :math:`\pi`.
    Om = Float(pi, desc='maximum solid angle')

    # internal identifier
    digest = Property(
        depends_on=['c', 'ff.digest', 'N', 'Om'],
    )

    # internal dictionary of interpolators
    idict = Dict

    @cached_property
    def _get_digest(self):
        return digest(self)

    def _r(self, gpos, mpos=0.0):
        """Calculates the virtual distances between grid point locations and
        microphone locations or the origin. These virtual distances correspond
        to travel times of the sound along a ray that is traced through the
        medium. Functionality may change in the future.

        Parameters
        ----------
        gpos : array of floats of shape (3, N)
            The locations of points in the beamforming map grid in 3D cartesian
            co-ordinates.
        mpos : array of floats of shape (3, M), optional
            The locations of microphones in 3D cartesian co-ordinates. If not
            given, then only one microphone at the origin (0, 0, 0) is
            considered.

        Returns
        -------
        array of floats
            The distances in a twodimensional (N, M) array of floats. If M==1,
            then only a one-dimensional array is returned.

        """
        c = self.c

        if isscalar(mpos):
            mpos = array((0, 0, 0), dtype=float32)[:, newaxis]

        gt = empty((gpos.shape[-1], mpos.shape[-1]))
        for micnum, x0 in enumerate(mpos.T):
            key = x0.tobytes()  # make array hashable
            try:
                li = self.idict[key]  # fetch stored interpolator
            except KeyError:
                # if interpolator doesn't exist, construct it
                roi = gpos
                if self.roi is not None:
                    roi = self.roi
                li = self.get_interpolator(roi, x0)
                self.idict[key] = li
            # do the interpolation
            gt[:, micnum] = li(gpos.T)
        if gt.shape[1] == 1:
            gt = gt[:, 0]
        return c * gt  # return distance along ray

    def get_interpolator(self, roi, x0):
        """Gets an LinearNDInterpolator object.

        Parameters
        ----------
        roi : array of floats of shape (3, N)
            The locations of points in the region of interest in 3D cartesian
            co-ordinates. Used to estimate the maximum distance and ROI
            extension and center.
        x0 : array of floats of shape (3)
            The location of the microphone in 3D cartesian co-ordinates.

        Returns
        -------
        LinearNDInterpolator object

        """
        c = self.c

        # the DE system
        def f1(t, y, v):  # noqa: ARG001
            x = y[0:3]
            s = y[3:6]
            vv, dv = v(x)
            sa = sqrt(s[0] * s[0] + s[1] * s[1] + s[2] * s[2])
            x = empty(6)
            x[0:3] = c * s / sa - vv  # time reversal
            x[3:6] = dot(s, -dv.T)  # time reversal
            return x

        # integration along a single ray
        def fr(x0, n0, rmax, dt, v, xyz, t):
            s0 = n0 / (c + dot(v(x0)[0], n0))
            y0 = hstack((x0, s0))
            oo = ode(f1)
            oo.set_f_params(v)
            oo.set_integrator(
                'vode',
                rtol=1e-4,  # accuracy !
                max_step=1e-4 * rmax,
            )  # for thin shear layer
            oo.set_initial_value(y0, 0)
            while oo.successful():
                xyz.append(oo.y[0:3])
                t.append(oo.t)
                if norm(oo.y[0:3] - x0) > rmax:
                    break
                oo.integrate(oo.t + dt)

        gs2 = roi.shape[-1]
        vv = self.ff.v
        NN = int(sqrt(self.N))
        xe = roi.mean(1)  # center of grid
        r = x0[:, newaxis] - roi
        rmax = sqrt((r * r).sum(0).max())  # maximum distance
        nv = spiral_sphere(self.N, self.Om, b=xe - x0)
        rstep = rmax / sqrt(self.N)
        rmax += rstep
        tstep = rstep / c
        xyz = []
        t = []
        lastind = 0
        for i, n0 in enumerate(nv.T):
            fr(x0, n0, rmax, tstep, vv, xyz, t)
            if i and i % NN == 0:
                if not lastind:
                    dd = ConvexHull(vstack((roi.T, xyz)), incremental=True)
                else:
                    dd.add_points(xyz[lastind:], restart=True)
                lastind = len(xyz)
                # ConvexHull includes grid if no grid points on hull
                if dd.simplices.min() >= gs2:
                    break
        xyz = array(xyz)
        t = array(t)
        return LinearNDInterpolator(xyz, t)



# ------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
# ------------------------------------------------------------------------------
"""Implements support for array microphone arrangements.

.. autosummary::
    :toctree: generated/

    MicGeom

"""

# imports from other packages
import xml.dom.minidom
from pathlib import Path

from numpy import array, average
from scipy.spatial.distance import cdist
from traits.api import (
    CArray,
    File,
    HasStrictTraits,
    List,
    Property,
    cached_property,
    on_trait_change,
)

# acoular imports
from .deprecation import deprecated_alias
from .internal import digest


@deprecated_alias({'mpos_tot': 'pos_total', 'mpos': 'pos', 'from_file': 'file'}, read_only=['mpos'])
class MicGeom(HasStrictTraits):
    """Provides the geometric arrangement of microphones in the array.

    The geometric arrangement of microphones is read in from an
    xml-source with element tag names `pos` and attributes Name, `x`, `y` and `z`.
    Can also be used with programmatically generated arrangements.
    """

    #: Name of the .xml-file from which to read the data.
    file = File(filter=['*.xml'], exists=True, desc='name of the xml file to import')

    #: Positions as (3, :attr:`num_mics`) array of floats, may include also invalid
    #: microphones (if any). Set either automatically on change of the
    #: :attr:`file` argument or explicitly by assigning an array of floats.
    pos_total = CArray(dtype=float, shape=(3, None), desc='x, y, z position of all microphones')

    #: Positions as (3, :attr:`num_mics`) array of floats, without invalid
    #: microphones; readonly.
    pos = Property(depends_on=['pos_total', 'invalid_channels'], desc='x, y, z position of used microphones')

    #: List that gives the indices of channels that should not be considered.
    #: Defaults to a blank list.
    invalid_channels = List(int, desc='list of invalid channels')

    #: Number of used microphones in the array; readonly.
    num_mics = Property(depends_on=['pos'], desc='number of microphones in the geometry')

    #: Center of the array (arithmetic mean of all used array positions); readonly.
    center = Property(depends_on=['pos'], desc='array center')

    #: Aperture of the array (greatest extent between two microphones); readonly.
    aperture = Property(depends_on=['pos'], desc='array aperture')

    # internal identifier
    digest = Property(depends_on=['pos'])

    @cached_property
    def _get_digest(self):
        return digest(self)

    @cached_property
    def _get_pos(self):
        if len(self.invalid_channels) == 0:
            return self.pos_total
        allr = [i for i in range(self.pos_total.shape[-1]) if i not in self.invalid_channels]
        return self.pos_total[:, array(allr)]

    @cached_property
    def _get_num_mics(self):
        return self.pos.shape[-1]

    @cached_property
    def _get_center(self):
        if self.pos.any():
            center = average(self.pos, axis=1)
            # set very small values to zero
            center[abs(center) < 1e-16] = 0.0
            return center
        return None

    @cached_property
    def _get_aperture(self):
        if self.pos.any():
            return cdist(self.pos.T, self.pos.T).max()
        return None

    @on_trait_change('file')
    def import_mpos(self):
        """Import the microphone positions from .xml file.
        Called when :attr:`file` changes.
        """
        doc = xml.dom.minidom.parse(self.file)
        names = []
        xyz = []
        for el in doc.getElementsByTagName('pos'):
            names.append(el.getAttribute('Name'))
            xyz.append([float(el.getAttribute(a)) for a in 'xyz'])
        self.pos_total = array(xyz, 'd').swapaxes(0, 1)

    def export_mpos(self, filename):
        """Export the microphone positions to .xml file.

        Parameters
        ----------
        filename : str
            Name of the file to which the microphone positions are written.
        """
        filepath = Path(filename)
        basename = filepath.stem
        with filepath.open('w', encoding='utf-8') as f:
            f.write(f'<?xml version="1.1" encoding="utf-8"?><MicArray name="{basename}">\n')
            for i in range(self.pos.shape[-1]):
                f.write(
                    f'  <pos Name="Point {i+1}" x="{self.pos[0, i]}" y="{self.pos[1, i]}" z="{self.pos[2, i]}"/>\n',
                )
            f.write('</MicArray>')
